# WARNING THIS FILE IS AUTOGENERATED
#
# Databricks CLI
# Copyright 2017 Databricks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"), except
# that the use of services to which certain application programming
# interfaces (each, an "API") connect requires that the user first obtain
# a license for the use of the APIs from Databricks, Inc. ("Databricks"),
# by creating an account at www.databricks.com and agreeing to either (a)
# the Community Edition Terms of Service, (b) the Databricks Terms of
# Service, or (c) another written agreement between Licensee and Databricks
# for the use of the APIs.
#
# You may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import os
import re

from six.moves.urllib.parse import urlparse


class JobsService(object):
    def __init__(self, client):
        self.client = client

    def create_job(
        self,
        name=None,
        existing_cluster_id=None,
        new_cluster=None,
        libraries=None,
        email_notifications=None,
        timeout_seconds=None,
        max_retries=None,
        min_retry_interval_millis=None,
        retry_on_timeout=None,
        schedule=None,
        notebook_task=None,
        spark_jar_task=None,
        spark_python_task=None,
        spark_submit_task=None,
        max_concurrent_runs=None,
        tasks=None,
        headers=None,
        version=None,
        git_source=None,
        job_clusters=None,
        tags=None,
        format=None,
        dbt_task=None,
        access_control_list=None,
        pipeline_task=None,
        python_wheel_task=None,
        sql_task=None,
        webhook_notifications=None,
        continuous=None,
    ):
        _data = {}
        if name is not None:
            _data['name'] = name
        if existing_cluster_id is not None:
            _data['existing_cluster_id'] = existing_cluster_id
        if new_cluster is not None:
            _data['new_cluster'] = new_cluster
            if not isinstance(new_cluster, dict):
                raise TypeError('Expected databricks.NewCluster() or dict for field new_cluster')
        if libraries is not None:
            _data['libraries'] = libraries
        if email_notifications is not None:
            _data['email_notifications'] = email_notifications
            if not isinstance(email_notifications, dict):
                raise TypeError(
                    'Expected databricks.JobEmailNotifications() or dict for field email_notifications'
                )
        if timeout_seconds is not None:
            _data['timeout_seconds'] = timeout_seconds
        if max_retries is not None:
            _data['max_retries'] = max_retries
        if min_retry_interval_millis is not None:
            _data['min_retry_interval_millis'] = min_retry_interval_millis
        if retry_on_timeout is not None:
            _data['retry_on_timeout'] = retry_on_timeout
        if schedule is not None:
            _data['schedule'] = schedule
            if not isinstance(schedule, dict):
                raise TypeError('Expected databricks.CronSchedule() or dict for field schedule')
        if notebook_task is not None:
            _data['notebook_task'] = notebook_task
            if not isinstance(notebook_task, dict):
                raise TypeError(
                    'Expected databricks.NotebookTask() or dict for field notebook_task'
                )
        if spark_jar_task is not None:
            _data['spark_jar_task'] = spark_jar_task
            if not isinstance(spark_jar_task, dict):
                raise TypeError(
                    'Expected databricks.SparkJarTask() or dict for field spark_jar_task'
                )
        if spark_python_task is not None:
            _data['spark_python_task'] = spark_python_task
            if not isinstance(spark_python_task, dict):
                raise TypeError(
                    'Expected databricks.SparkPythonTask() or dict for field spark_python_task'
                )
        if spark_submit_task is not None:
            _data['spark_submit_task'] = spark_submit_task
            if not isinstance(spark_submit_task, dict):
                raise TypeError(
                    'Expected databricks.SparkSubmitTask() or dict for field spark_submit_task'
                )
        if max_concurrent_runs is not None:
            _data['max_concurrent_runs'] = max_concurrent_runs
        if tasks is not None:
            _data['tasks'] = tasks
        if git_source is not None:
            _data['git_source'] = git_source
            if not isinstance(git_source, dict):
                raise TypeError('Expected databricks.GitSource() or dict for field git_source')
        if job_clusters is not None:
            _data['job_clusters'] = job_clusters
        if tags is not None:
            _data['tags'] = tags
        if format is not None:
            _data['format'] = format
        if dbt_task is not None:
            _data['dbt_task'] = dbt_task
            if not isinstance(dbt_task, dict):
                raise TypeError('Expected databricks.DbtTask() or dict for field dbt_task')
        if access_control_list is not None:
            _data['access_control_list'] = access_control_list
        if pipeline_task is not None:
            _data['pipeline_task'] = pipeline_task
            if not isinstance(pipeline_task, dict):
                raise TypeError(
                    'Expected databricks.PipelineTask() or dict for field pipeline_task'
                )
        if python_wheel_task is not None:
            _data['python_wheel_task'] = python_wheel_task
            if not isinstance(python_wheel_task, dict):
                raise TypeError(
                    'Expected databricks.PythonWheelTask() or dict for field python_wheel_task'
                )
        if sql_task is not None:
            _data['sql_task'] = sql_task
            if not isinstance(sql_task, dict):
                raise TypeError('Expected databricks.SqlTask() or dict for field sql_task')
        if webhook_notifications is not None:
            _data['webhook_notifications'] = webhook_notifications
            if not isinstance(webhook_notifications, dict):
                raise TypeError(
                    'Expected databricks.WebhookNotifications() or dict for field webhook_notifications'
                )
        if continuous is not None:
            _data['continuous'] = continuous
            if not isinstance(continuous, dict):
                raise TypeError(
                    'Expected databricks.ContinuousSettings() or dict for field continuous'
                )
        return self.client.perform_query(
            'POST', '/jobs/create', data=_data, headers=headers, version=version
        )

    def submit_run(
        self,
        run_name=None,
        existing_cluster_id=None,
        new_cluster=None,
        libraries=None,
        notebook_task=None,
        spark_jar_task=None,
        spark_python_task=None,
        spark_submit_task=None,
        timeout_seconds=None,
        tasks=None,
        headers=None,
        version=None,
        idempotency_token=None,
        job_clusters=None,
        git_source=None,
        dbt_task=None,
        access_control_list=None,
        pipeline_task=None,
        python_wheel_task=None,
        sql_task=None,
        webhook_notifications=None,
    ):
        _data = {}
        if run_name is not None:
            _data['run_name'] = run_name
        if existing_cluster_id is not None:
            _data['existing_cluster_id'] = existing_cluster_id
        if new_cluster is not None:
            _data['new_cluster'] = new_cluster
            if not isinstance(new_cluster, dict):
                raise TypeError('Expected databricks.NewCluster() or dict for field new_cluster')
        if libraries is not None:
            _data['libraries'] = libraries
        if notebook_task is not None:
            _data['notebook_task'] = notebook_task
            if not isinstance(notebook_task, dict):
                raise TypeError(
                    'Expected databricks.NotebookTask() or dict for field notebook_task'
                )
        if spark_jar_task is not None:
            _data['spark_jar_task'] = spark_jar_task
            if not isinstance(spark_jar_task, dict):
                raise TypeError(
                    'Expected databricks.SparkJarTask() or dict for field spark_jar_task'
                )
        if spark_python_task is not None:
            _data['spark_python_task'] = spark_python_task
            if not isinstance(spark_python_task, dict):
                raise TypeError(
                    'Expected databricks.SparkPythonTask() or dict for field spark_python_task'
                )
        if spark_submit_task is not None:
            _data['spark_submit_task'] = spark_submit_task
            if not isinstance(spark_submit_task, dict):
                raise TypeError(
                    'Expected databricks.SparkSubmitTask() or dict for field spark_submit_task'
                )
        if timeout_seconds is not None:
            _data['timeout_seconds'] = timeout_seconds
        if tasks is not None:
            _data['tasks'] = tasks
        if idempotency_token is not None:
            _data['idempotency_token'] = idempotency_token
        if job_clusters is not None:
            _data['job_clusters'] = job_clusters
        if git_source is not None:
            _data['git_source'] = git_source
            if not isinstance(git_source, dict):
                raise TypeError('Expected databricks.GitSource() or dict for field git_source')
        if dbt_task is not None:
            _data['dbt_task'] = dbt_task
            if not isinstance(dbt_task, dict):
                raise TypeError('Expected databricks.DbtTask() or dict for field dbt_task')
        if access_control_list is not None:
            _data['access_control_list'] = access_control_list
        if pipeline_task is not None:
            _data['pipeline_task'] = pipeline_task
            if not isinstance(pipeline_task, dict):
                raise TypeError(
                    'Expected databricks.PipelineTask() or dict for field pipeline_task'
                )
        if python_wheel_task is not None:
            _data['python_wheel_task'] = python_wheel_task
            if not isinstance(python_wheel_task, dict):
                raise TypeError(
                    'Expected databricks.PythonWheelTask() or dict for field python_wheel_task'
                )
        if sql_task is not None:
            _data['sql_task'] = sql_task
            if not isinstance(sql_task, dict):
                raise TypeError('Expected databricks.SqlTask() or dict for field sql_task')
        if webhook_notifications is not None:
            _data['webhook_notifications'] = webhook_notifications
            if not isinstance(webhook_notifications, dict):
                raise TypeError(
                    'Expected databricks.WebhookNotifications() or dict for field webhook_notifications'
                )
        return self.client.perform_query(
            'POST', '/jobs/runs/submit', data=_data, headers=headers, version=version
        )

    def reset_job(self, job_id, new_settings, headers=None, version=None):
        _data = {}
        if job_id is not None:
            _data['job_id'] = job_id
        if new_settings is not None:
            _data['new_settings'] = new_settings
            if not isinstance(new_settings, dict):
                raise TypeError('Expected databricks.JobSettings() or dict for field new_settings')
        return self.client.perform_query(
            'POST', '/jobs/reset', data=_data, headers=headers, version=version
        )

    def update_job(
        self, job_id, new_settings=None, fields_to_remove=None, headers=None, version=None
    ):
        _data = {}
        if job_id is not None:
            _data['job_id'] = job_id
        if new_settings is not None:
            _data['new_settings'] = new_settings
            if not isinstance(new_settings, dict):
                raise TypeError('Expected databricks.JobSettings() or dict for field new_settings')
        if fields_to_remove is not None:
            _data['fields_to_remove'] = fields_to_remove
        return self.client.perform_query(
            'POST', '/jobs/update', data=_data, headers=headers, version=version
        )

    def delete_job(self, job_id, headers=None, version=None):
        _data = {}
        if job_id is not None:
            _data['job_id'] = job_id
        return self.client.perform_query(
            'POST', '/jobs/delete', data=_data, headers=headers, version=version
        )

    def get_job(self, job_id, headers=None, version=None, include_trigger_history=None):
        _data = {}
        if job_id is not None:
            _data['job_id'] = job_id
        if include_trigger_history is not None:
            _data['include_trigger_history'] = include_trigger_history
        return self.client.perform_query(
            'GET', '/jobs/get', data=_data, headers=headers, version=version
        )

    def list_jobs(
        self,
        job_type=None,
        expand_tasks=None,
        limit=None,
        offset=None,
        headers=None,
        version=None,
        name=None,
    ):
        _data = {}
        if expand_tasks is not None:
            _data['expand_tasks'] = expand_tasks
        if limit is not None:
            _data['limit'] = limit
        if offset is not None:
            _data['offset'] = offset
        if name is not None:
            _data['name'] = name
        return self.client.perform_query(
            'GET', '/jobs/list', data=_data, headers=headers, version=version
        )

    def run_now(
        self,
        job_id=None,
        jar_params=None,
        notebook_params=None,
        python_params=None,
        spark_submit_params=None,
        python_named_params=None,
        idempotency_token=None,
        headers=None,
        version=None,
        dbt_commands=None,
        pipeline_params=None,
    ):
        _data = {}
        if job_id is not None:
            _data['job_id'] = job_id
        if jar_params is not None:
            _data['jar_params'] = jar_params
        if notebook_params is not None:
            _data['notebook_params'] = notebook_params
        if python_params is not None:
            _data['python_params'] = python_params
        if spark_submit_params is not None:
            _data['spark_submit_params'] = spark_submit_params
        if python_named_params is not None:
            _data['python_named_params'] = python_named_params
        if idempotency_token is not None:
            _data['idempotency_token'] = idempotency_token
        if dbt_commands is not None:
            _data['dbt_commands'] = dbt_commands
        if pipeline_params is not None:
            _data['pipeline_params'] = pipeline_params
            if not isinstance(pipeline_params, dict):
                raise TypeError(
                    'Expected databricks.PipelineParameters() or dict for field pipeline_params'
                )
        return self.client.perform_query(
            'POST', '/jobs/run-now', data=_data, headers=headers, version=version
        )

    def repair(
        self,
        run_id,
        latest_repair_id=None,
        rerun_tasks=None,
        jar_params=None,
        notebook_params=None,
        python_params=None,
        spark_submit_params=None,
        python_named_params=None,
        headers=None,
        version=None,
        dbt_commands=None,
        pipeline_params=None,
        rerun_all_failed_tasks=None,
        rerun_dependent_tasks=None,
    ):
        _data = {}
        if run_id is not None:
            _data['run_id'] = run_id
        if latest_repair_id is not None:
            _data['latest_repair_id'] = latest_repair_id
        if rerun_tasks is not None:
            _data['rerun_tasks'] = rerun_tasks
        if jar_params is not None:
            _data['jar_params'] = jar_params
        if notebook_params is not None:
            _data['notebook_params'] = notebook_params
        if python_params is not None:
            _data['python_params'] = python_params
        if spark_submit_params is not None:
            _data['spark_submit_params'] = spark_submit_params
        if python_named_params is not None:
            _data['python_named_params'] = python_named_params
        if dbt_commands is not None:
            _data['dbt_commands'] = dbt_commands
        if pipeline_params is not None:
            _data['pipeline_params'] = pipeline_params
            if not isinstance(pipeline_params, dict):
                raise TypeError(
                    'Expected databricks.PipelineParameters() or dict for field pipeline_params'
                )
        if rerun_all_failed_tasks is not None:
            _data['rerun_all_failed_tasks'] = rerun_all_failed_tasks
        if rerun_dependent_tasks is not None:
            _data['rerun_dependent_tasks'] = rerun_dependent_tasks
        return self.client.perform_query(
            'POST', '/jobs/runs/repair', data=_data, headers=headers, version=version
        )

    def list_runs(
        self,
        job_id=None,
        active_only=None,
        completed_only=None,
        offset=None,
        limit=None,
        headers=None,
        version=None,
        run_type=None,
        expand_tasks=None,
        start_time_from=None,
        start_time_to=None,
        page_token=None,
    ):
        _data = {}
        if job_id is not None:
            _data['job_id'] = job_id
        if active_only is not None:
            _data['active_only'] = active_only
        if completed_only is not None:
            _data['completed_only'] = completed_only
        if offset is not None:
            _data['offset'] = offset
        if limit is not None:
            _data['limit'] = limit
        if run_type is not None:
            _data['run_type'] = run_type
        if expand_tasks is not None:
            _data['expand_tasks'] = expand_tasks
        if start_time_from is not None:
            _data['start_time_from'] = start_time_from
        if start_time_to is not None:
            _data['start_time_to'] = start_time_to
        if page_token is not None:
            _data['page_token'] = page_token
        return self.client.perform_query(
            'GET', '/jobs/runs/list', data=_data, headers=headers, version=version
        )

    def get_run(self, run_id=None, headers=None, version=None, include_history=None):
        _data = {}
        if run_id is not None:
            _data['run_id'] = run_id
        if include_history is not None:
            _data['include_history'] = include_history
        return self.client.perform_query(
            'GET', '/jobs/runs/get', data=_data, headers=headers, version=version
        )

    def delete_run(self, run_id=None, headers=None, version=None):
        _data = {}
        if run_id is not None:
            _data['run_id'] = run_id
        return self.client.perform_query(
            'POST', '/jobs/runs/delete', data=_data, headers=headers, version=version
        )

    def cancel_run(self, run_id, headers=None, version=None):
        _data = {}
        if run_id is not None:
            _data['run_id'] = run_id
        return self.client.perform_query(
            'POST', '/jobs/runs/cancel', data=_data, headers=headers, version=version
        )

    def cancel_all_runs(self, job_id=None, headers=None, version=None, all_queued_runs=None):
        _data = {}
        if job_id is not None:
            _data['job_id'] = job_id
        if all_queued_runs is not None:
            _data['all_queued_runs'] = all_queued_runs
        return self.client.perform_query(
            'POST', '/jobs/runs/cancel-all', data=_data, headers=headers, version=version
        )

    def get_run_output(self, run_id, headers=None, version=None):
        _data = {}
        if run_id is not None:
            _data['run_id'] = run_id
        return self.client.perform_query(
            'GET', '/jobs/runs/get-output', data=_data, headers=headers, version=version
        )

    def export_run(self, run_id, views_to_export=None, headers=None, version=None):
        _data = {}
        if run_id is not None:
            _data['run_id'] = run_id
        if views_to_export is not None:
            _data['views_to_export'] = views_to_export
        return self.client.perform_query(
            'GET', '/jobs/runs/export', data=_data, headers=headers, version=version
        )


class ClusterService(object):
    def __init__(self, client):
        self.client = client

    def list_clusters(self, headers=None):
        _data = {}
        return self.client.perform_query('GET', '/clusters/list', data=_data, headers=headers)

    def create_cluster(
        self,
        num_workers=None,
        autoscale=None,
        cluster_name=None,
        spark_version=None,
        spark_conf=None,
        aws_attributes=None,
        node_type_id=None,
        driver_node_type_id=None,
        ssh_public_keys=None,
        custom_tags=None,
        cluster_log_conf=None,
        spark_env_vars=None,
        autotermination_minutes=None,
        enable_elastic_disk=None,
        cluster_source=None,
        instance_pool_id=None,
        headers=None,
        azure_attributes=None,
        gcp_attributes=None,
        policy_id=None,
        enable_local_disk_encryption=None,
        driver_instance_pool_id=None,
        apply_policy_default_values=None,
    ):
        _data = {}
        if num_workers is not None:
            _data['num_workers'] = num_workers
        if autoscale is not None:
            _data['autoscale'] = autoscale
            if not isinstance(autoscale, dict):
                raise TypeError('Expected databricks.AutoScale() or dict for field autoscale')
        if cluster_name is not None:
            _data['cluster_name'] = cluster_name
        if spark_version is not None:
            _data['spark_version'] = spark_version
        if spark_conf is not None:
            _data['spark_conf'] = spark_conf
        if aws_attributes is not None:
            _data['aws_attributes'] = aws_attributes
            if not isinstance(aws_attributes, dict):
                raise TypeError(
                    'Expected databricks.AwsAttributes() or dict for field aws_attributes'
                )
        if node_type_id is not None:
            _data['node_type_id'] = node_type_id
        if driver_node_type_id is not None:
            _data['driver_node_type_id'] = driver_node_type_id
        if ssh_public_keys is not None:
            _data['ssh_public_keys'] = ssh_public_keys
        if custom_tags is not None:
            _data['custom_tags'] = custom_tags
        if cluster_log_conf is not None:
            _data['cluster_log_conf'] = cluster_log_conf
            if not isinstance(cluster_log_conf, dict):
                raise TypeError(
                    'Expected databricks.ClusterLogConf() or dict for field cluster_log_conf'
                )
        if spark_env_vars is not None:
            _data['spark_env_vars'] = spark_env_vars
        if autotermination_minutes is not None:
            _data['autotermination_minutes'] = autotermination_minutes
        if enable_elastic_disk is not None:
            _data['enable_elastic_disk'] = enable_elastic_disk
        if cluster_source is not None:
            _data['cluster_source'] = cluster_source
        if instance_pool_id is not None:
            _data['instance_pool_id'] = instance_pool_id
        if azure_attributes is not None:
            _data['azure_attributes'] = azure_attributes
            if not isinstance(azure_attributes, dict):
                raise TypeError(
                    'Expected databricks.AzureAttributes() or dict for field azure_attributes'
                )
        if gcp_attributes is not None:
            _data['gcp_attributes'] = gcp_attributes
            if not isinstance(gcp_attributes, dict):
                raise TypeError(
                    'Expected databricks.GcpAttributes() or dict for field gcp_attributes'
                )
        if policy_id is not None:
            _data['policy_id'] = policy_id
        if enable_local_disk_encryption is not None:
            _data['enable_local_disk_encryption'] = enable_local_disk_encryption
        if driver_instance_pool_id is not None:
            _data['driver_instance_pool_id'] = driver_instance_pool_id
        if apply_policy_default_values is not None:
            _data['apply_policy_default_values'] = apply_policy_default_values
        return self.client.perform_query('POST', '/clusters/create', data=_data, headers=headers)

    def start_cluster(self, cluster_id, headers=None):
        _data = {}
        if cluster_id is not None:
            _data['cluster_id'] = cluster_id
        return self.client.perform_query('POST', '/clusters/start', data=_data, headers=headers)

    def list_spark_versions(self, headers=None):
        _data = {}
        return self.client.perform_query(
            'GET', '/clusters/spark-versions', data=_data, headers=headers
        )

    def delete_cluster(self, cluster_id, headers=None):
        _data = {}
        if cluster_id is not None:
            _data['cluster_id'] = cluster_id
        return self.client.perform_query('POST', '/clusters/delete', data=_data, headers=headers)

    def permanent_delete_cluster(self, cluster_id, headers=None):
        _data = {}
        if cluster_id is not None:
            _data['cluster_id'] = cluster_id
        return self.client.perform_query(
            'POST', '/clusters/permanent-delete', data=_data, headers=headers
        )

    def restart_cluster(self, cluster_id, headers=None):
        _data = {}
        if cluster_id is not None:
            _data['cluster_id'] = cluster_id
        return self.client.perform_query('POST', '/clusters/restart', data=_data, headers=headers)

    def resize_cluster(self, cluster_id, num_workers=None, autoscale=None, headers=None):
        _data = {}
        if cluster_id is not None:
            _data['cluster_id'] = cluster_id
        if num_workers is not None:
            _data['num_workers'] = num_workers
        if autoscale is not None:
            _data['autoscale'] = autoscale
            if not isinstance(autoscale, dict):
                raise TypeError('Expected databricks.AutoScale() or dict for field autoscale')
        return self.client.perform_query('POST', '/clusters/resize', data=_data, headers=headers)

    def edit_cluster(
        self,
        cluster_id,
        num_workers=None,
        autoscale=None,
        cluster_name=None,
        spark_version=None,
        spark_conf=None,
        aws_attributes=None,
        node_type_id=None,
        driver_node_type_id=None,
        ssh_public_keys=None,
        custom_tags=None,
        cluster_log_conf=None,
        spark_env_vars=None,
        autotermination_minutes=None,
        enable_elastic_disk=None,
        cluster_source=None,
        instance_pool_id=None,
        headers=None,
        azure_attributes=None,
        gcp_attributes=None,
        policy_id=None,
        enable_local_disk_encryption=None,
        driver_instance_pool_id=None,
        apply_policy_default_values=None,
    ):
        _data = {}
        if cluster_id is not None:
            _data['cluster_id'] = cluster_id
        if num_workers is not None:
            _data['num_workers'] = num_workers
        if autoscale is not None:
            _data['autoscale'] = autoscale
            if not isinstance(autoscale, dict):
                raise TypeError('Expected databricks.AutoScale() or dict for field autoscale')
        if cluster_name is not None:
            _data['cluster_name'] = cluster_name
        if spark_version is not None:
            _data['spark_version'] = spark_version
        if spark_conf is not None:
            _data['spark_conf'] = spark_conf
        if aws_attributes is not None:
            _data['aws_attributes'] = aws_attributes
            if not isinstance(aws_attributes, dict):
                raise TypeError(
                    'Expected databricks.AwsAttributes() or dict for field aws_attributes'
                )
        if node_type_id is not None:
            _data['node_type_id'] = node_type_id
        if driver_node_type_id is not None:
            _data['driver_node_type_id'] = driver_node_type_id
        if ssh_public_keys is not None:
            _data['ssh_public_keys'] = ssh_public_keys
        if custom_tags is not None:
            _data['custom_tags'] = custom_tags
        if cluster_log_conf is not None:
            _data['cluster_log_conf'] = cluster_log_conf
            if not isinstance(cluster_log_conf, dict):
                raise TypeError(
                    'Expected databricks.ClusterLogConf() or dict for field cluster_log_conf'
                )
        if spark_env_vars is not None:
            _data['spark_env_vars'] = spark_env_vars
        if autotermination_minutes is not None:
            _data['autotermination_minutes'] = autotermination_minutes
        if enable_elastic_disk is not None:
            _data['enable_elastic_disk'] = enable_elastic_disk
        if cluster_source is not None:
            _data['cluster_source'] = cluster_source
        if instance_pool_id is not None:
            _data['instance_pool_id'] = instance_pool_id
        if azure_attributes is not None:
            _data['azure_attributes'] = azure_attributes
            if not isinstance(azure_attributes, dict):
                raise TypeError(
                    'Expected databricks.AzureAttributes() or dict for field azure_attributes'
                )
        if gcp_attributes is not None:
            _data['gcp_attributes'] = gcp_attributes
            if not isinstance(gcp_attributes, dict):
                raise TypeError(
                    'Expected databricks.GcpAttributes() or dict for field gcp_attributes'
                )
        if policy_id is not None:
            _data['policy_id'] = policy_id
        if enable_local_disk_encryption is not None:
            _data['enable_local_disk_encryption'] = enable_local_disk_encryption
        if driver_instance_pool_id is not None:
            _data['driver_instance_pool_id'] = driver_instance_pool_id
        if apply_policy_default_values is not None:
            _data['apply_policy_default_values'] = apply_policy_default_values
        return self.client.perform_query('POST', '/clusters/edit', data=_data, headers=headers)

    def get_cluster(self, cluster_id, headers=None):
        _data = {}
        if cluster_id is not None:
            _data['cluster_id'] = cluster_id
        return self.client.perform_query('GET', '/clusters/get', data=_data, headers=headers)

    def pin_cluster(self, cluster_id, headers=None):
        _data = {}
        if cluster_id is not None:
            _data['cluster_id'] = cluster_id
        return self.client.perform_query('POST', '/clusters/pin', data=_data, headers=headers)

    def unpin_cluster(self, cluster_id, headers=None):
        _data = {}
        if cluster_id is not None:
            _data['cluster_id'] = cluster_id
        return self.client.perform_query('POST', '/clusters/unpin', data=_data, headers=headers)

    def list_node_types(self, headers=None):
        _data = {}
        return self.client.perform_query(
            'GET', '/clusters/list-node-types', data=_data, headers=headers
        )

    def list_available_zones(self, headers=None):
        _data = {}
        return self.client.perform_query('GET', '/clusters/list-zones', data=_data, headers=headers)

    def get_events(
        self,
        cluster_id,
        start_time=None,
        end_time=None,
        order=None,
        event_types=None,
        offset=None,
        limit=None,
        headers=None,
    ):
        _data = {}
        if cluster_id is not None:
            _data['cluster_id'] = cluster_id
        if start_time is not None:
            _data['start_time'] = start_time
        if end_time is not None:
            _data['end_time'] = end_time
        if order is not None:
            _data['order'] = order
        if event_types is not None:
            _data['event_types'] = event_types
        if offset is not None:
            _data['offset'] = offset
        if limit is not None:
            _data['limit'] = limit
        return self.client.perform_query('POST', '/clusters/events', data=_data, headers=headers)


class PolicyService(object):
    def __init__(self, client):
        self.client = client

    def get_policy(self, policy_id=None, headers=None):
        _data = {}
        if policy_id is not None:
            _data['policy_id'] = policy_id
        return self.client.perform_query(
            'GET', '/policies/clusters/get', data=_data, headers=headers
        )

    def list_policies(self, headers=None):
        _data = {}
        return self.client.perform_query(
            'GET', '/policies/clusters/list', data=_data, headers=headers
        )

    def create_policy(self, policy_name=None, definition=None, headers=None, name=None):
        _data = {}
        if definition is not None:
            _data['definition'] = definition
        if name is not None:
            _data['name'] = name
        return self.client.perform_query(
            'POST', '/policies/clusters/create', data=_data, headers=headers
        )

    def edit_policy(self, policy_id, policy_name=None, definition=None, headers=None, name=None):
        _data = {}
        if policy_id is not None:
            _data['policy_id'] = policy_id
        if definition is not None:
            _data['definition'] = definition
        if name is not None:
            _data['name'] = name
        return self.client.perform_query(
            'POST', '/policies/clusters/edit', data=_data, headers=headers
        )

    def delete_policy(self, policy_id, headers=None):
        _data = {}
        if policy_id is not None:
            _data['policy_id'] = policy_id
        return self.client.perform_query(
            'POST', '/policies/clusters/delete', data=_data, headers=headers
        )


class ManagedLibraryService(object):
    def __init__(self, client):
        self.client = client

    def cluster_status(self, cluster_id, headers=None):
        _data = {}
        if cluster_id is not None:
            _data['cluster_id'] = cluster_id
        return self.client.perform_query(
            'GET', '/libraries/cluster-status', data=_data, headers=headers
        )

    def all_cluster_statuses(self, headers=None):
        _data = {}
        return self.client.perform_query(
            'GET', '/libraries/all-cluster-statuses', data=_data, headers=headers
        )

    def install_libraries(self, cluster_id, libraries=None, headers=None):
        _data = {}
        if cluster_id is not None:
            _data['cluster_id'] = cluster_id
        if libraries is not None:
            _data['libraries'] = libraries
        return self.client.perform_query('POST', '/libraries/install', data=_data, headers=headers)

    def uninstall_libraries(self, cluster_id, libraries=None, headers=None):
        _data = {}
        if cluster_id is not None:
            _data['cluster_id'] = cluster_id
        if libraries is not None:
            _data['libraries'] = libraries
        return self.client.perform_query(
            'POST', '/libraries/uninstall', data=_data, headers=headers
        )


class DbfsService(object):
    def __init__(self, client):
        self.client = client

    def read(self, path, offset=None, length=None, headers=None):
        _data = {}
        if path is not None:
            _data['path'] = path
        if offset is not None:
            _data['offset'] = offset
        if length is not None:
            _data['length'] = length
        return self.client.perform_query('GET', '/dbfs/read', data=_data, headers=headers)

    def get_status(self, path, headers=None):
        _data = {}
        if path is not None:
            _data['path'] = path
        return self.client.perform_query('GET', '/dbfs/get-status', data=_data, headers=headers)

    def list(self, path, headers=None):
        _data = {}
        if path is not None:
            _data['path'] = path
        return self.client.perform_query('GET', '/dbfs/list', data=_data, headers=headers)

    def put(self, path, contents=None, overwrite=None, headers=None, src_path=None):
        _data = {}
        _files = None
        if path is not None:
            _data['path'] = path
        if contents is not None:
            _data['contents'] = contents
        if overwrite is not None:
            _data['overwrite'] = overwrite
        if src_path is not None:
            headers = {'Content-Type': None}
            filename = os.path.basename(src_path)
            _files = {'file': (filename, open(src_path, 'rb'), 'multipart/form-data')}
        return self.client.perform_query(
            'POST', '/dbfs/put', data=_data, headers=headers, files=_files
        )

    def mkdirs(self, path, headers=None):
        _data = {}
        if path is not None:
            _data['path'] = path
        return self.client.perform_query('POST', '/dbfs/mkdirs', data=_data, headers=headers)

    def move(self, source_path, destination_path, headers=None):
        _data = {}
        if source_path is not None:
            _data['source_path'] = source_path
        if destination_path is not None:
            _data['destination_path'] = destination_path
        return self.client.perform_query('POST', '/dbfs/move', data=_data, headers=headers)

    def delete(self, path, recursive=None, headers=None):
        _data = {}
        if path is not None:
            _data['path'] = path
        if recursive is not None:
            _data['recursive'] = recursive
        return self.client.perform_query('POST', '/dbfs/delete', data=_data, headers=headers)

    def create(self, path, overwrite=None, headers=None):
        _data = {}
        if path is not None:
            _data['path'] = path
        if overwrite is not None:
            _data['overwrite'] = overwrite
        return self.client.perform_query('POST', '/dbfs/create', data=_data, headers=headers)

    def add_block(self, handle, data, headers=None):
        _data = {}
        if handle is not None:
            _data['handle'] = handle
        if data is not None:
            _data['data'] = data
        return self.client.perform_query('POST', '/dbfs/add-block', data=_data, headers=headers)

    def close(self, handle, headers=None):
        _data = {}
        if handle is not None:
            _data['handle'] = handle
        return self.client.perform_query('POST', '/dbfs/close', data=_data, headers=headers)


class WorkspaceService(object):
    def __init__(self, client):
        self.client = client

    def mkdirs(self, path, headers=None):
        _data = {}
        if path is not None:
            _data['path'] = path
        return self.client.perform_query('POST', '/workspace/mkdirs', data=_data, headers=headers)

    def list(self, path, headers=None):
        _data = {}
        if path is not None:
            _data['path'] = path
        return self.client.perform_query('GET', '/workspace/list', data=_data, headers=headers)

    def import_workspace(
        self, path, format=None, language=None, content=None, overwrite=None, headers=None
    ):
        _data = {}
        if path is not None:
            _data['path'] = path
        if format is not None:
            _data['format'] = format
        if language is not None:
            _data['language'] = language
        if content is not None:
            _data['content'] = content
        if overwrite is not None:
            _data['overwrite'] = overwrite
        return self.client.perform_query('POST', '/workspace/import', data=_data, headers=headers)

    def export_workspace(self, path, format=None, direct_download=None, headers=None):
        _data = {}
        if path is not None:
            _data['path'] = path
        if format is not None:
            _data['format'] = format
        if direct_download is not None:
            _data['direct_download'] = direct_download
        return self.client.perform_query('GET', '/workspace/export', data=_data, headers=headers)

    def delete(self, path, recursive=None, headers=None):
        _data = {}
        if path is not None:
            _data['path'] = path
        if recursive is not None:
            _data['recursive'] = recursive
        return self.client.perform_query('POST', '/workspace/delete', data=_data, headers=headers)

    def get_status(self, path, headers=None):
        _data = {}
        if path is not None:
            _data['path'] = path
        return self.client.perform_query(
            'GET', '/workspace/get-status', data=_data, headers=headers
        )


class SecretService(object):
    def __init__(self, client):
        self.client = client

    def create_scope(
        self,
        scope,
        initial_manage_principal=None,
        scope_backend_type=None,
        backend_azure_keyvault=None,
        headers=None,
    ):
        _data = {}
        if scope is not None:
            _data['scope'] = scope
        if initial_manage_principal is not None:
            _data['initial_manage_principal'] = initial_manage_principal
        if scope_backend_type is not None:
            _data['scope_backend_type'] = scope_backend_type
        if backend_azure_keyvault is not None:
            _data['backend_azure_keyvault'] = backend_azure_keyvault
            if not isinstance(backend_azure_keyvault, dict):
                raise TypeError(
                    'Expected databricks.AzureKeyVaultSecretScopeMetadata() or dict for field backend_azure_keyvault'
                )
        return self.client.perform_query(
            'POST', '/secrets/scopes/create', data=_data, headers=headers
        )

    def delete_scope(self, scope, headers=None):
        _data = {}
        if scope is not None:
            _data['scope'] = scope
        return self.client.perform_query(
            'POST', '/secrets/scopes/delete', data=_data, headers=headers
        )

    def list_scopes(self, headers=None):
        _data = {}
        return self.client.perform_query('GET', '/secrets/scopes/list', data=_data, headers=headers)

    def put_secret(self, scope, key, string_value=None, bytes_value=None, headers=None):
        _data = {}
        if scope is not None:
            _data['scope'] = scope
        if key is not None:
            _data['key'] = key
        if string_value is not None:
            _data['string_value'] = string_value
        if bytes_value is not None:
            _data['bytes_value'] = bytes_value
        return self.client.perform_query('POST', '/secrets/put', data=_data, headers=headers)

    def delete_secret(self, scope, key, headers=None):
        _data = {}
        if scope is not None:
            _data['scope'] = scope
        if key is not None:
            _data['key'] = key
        return self.client.perform_query('POST', '/secrets/delete', data=_data, headers=headers)

    def list_secrets(self, scope, headers=None):
        _data = {}
        if scope is not None:
            _data['scope'] = scope
        return self.client.perform_query('GET', '/secrets/list', data=_data, headers=headers)

    def put_acl(self, scope, principal, permission, headers=None):
        _data = {}
        if scope is not None:
            _data['scope'] = scope
        if principal is not None:
            _data['principal'] = principal
        if permission is not None:
            _data['permission'] = permission
        return self.client.perform_query('POST', '/secrets/acls/put', data=_data, headers=headers)

    def delete_acl(self, scope, principal, headers=None):
        _data = {}
        if scope is not None:
            _data['scope'] = scope
        if principal is not None:
            _data['principal'] = principal
        return self.client.perform_query(
            'POST', '/secrets/acls/delete', data=_data, headers=headers
        )

    def list_acls(self, scope, headers=None):
        _data = {}
        if scope is not None:
            _data['scope'] = scope
        return self.client.perform_query('GET', '/secrets/acls/list', data=_data, headers=headers)

    def get_acl(self, scope, principal, headers=None):
        _data = {}
        if scope is not None:
            _data['scope'] = scope
        if principal is not None:
            _data['principal'] = principal
        return self.client.perform_query('GET', '/secrets/acls/get', data=_data, headers=headers)


class GroupsService(object):
    def __init__(self, client):
        self.client = client

    def create_group(self, group_name, headers=None):
        _data = {}
        if group_name is not None:
            _data['group_name'] = group_name
        return self.client.perform_query('POST', '/groups/create', data=_data, headers=headers)

    def add_to_group(self, parent_name, user_name=None, group_name=None, headers=None):
        _data = {}
        if parent_name is not None:
            _data['parent_name'] = parent_name
        if user_name is not None:
            _data['user_name'] = user_name
        if group_name is not None:
            _data['group_name'] = group_name
        return self.client.perform_query('POST', '/groups/add-member', data=_data, headers=headers)

    def remove_from_group(self, parent_name, user_name=None, group_name=None, headers=None):
        _data = {}
        if parent_name is not None:
            _data['parent_name'] = parent_name
        if user_name is not None:
            _data['user_name'] = user_name
        if group_name is not None:
            _data['group_name'] = group_name
        return self.client.perform_query(
            'POST', '/groups/remove-member', data=_data, headers=headers
        )

    def get_groups(self, headers=None):
        _data = {}
        return self.client.perform_query('GET', '/groups/list', data=_data, headers=headers)

    def get_group_members(self, group_name, headers=None):
        _data = {}
        if group_name is not None:
            _data['group_name'] = group_name
        return self.client.perform_query('GET', '/groups/list-members', data=_data, headers=headers)

    def remove_group(self, group_name, headers=None):
        _data = {}
        if group_name is not None:
            _data['group_name'] = group_name
        return self.client.perform_query('POST', '/groups/delete', data=_data, headers=headers)

    def get_groups_for_principal(self, user_name=None, group_name=None, headers=None):
        _data = {}
        if user_name is not None:
            _data['user_name'] = user_name
        if group_name is not None:
            _data['group_name'] = group_name
        return self.client.perform_query('GET', '/groups/list-parents', data=_data, headers=headers)


class TokenService(object):
    def __init__(self, client):
        self.client = client

    def create_token(self, lifetime_seconds=None, comment=None, headers=None):
        _data = {}
        if lifetime_seconds is not None:
            _data['lifetime_seconds'] = lifetime_seconds
        if comment is not None:
            _data['comment'] = comment
        return self.client.perform_query('POST', '/token/create', data=_data, headers=headers)

    def list_tokens(self, headers=None):
        _data = {}
        return self.client.perform_query('GET', '/token/list', data=_data, headers=headers)

    def revoke_token(self, token_id, headers=None):
        _data = {}
        if token_id is not None:
            _data['token_id'] = token_id
        return self.client.perform_query('POST', '/token/delete', data=_data, headers=headers)


class InstancePoolService(object):
    def __init__(self, client):
        self.client = client

    def create_instance_pool(
        self,
        instance_pool_name=None,
        min_idle_instances=None,
        max_capacity=None,
        aws_attributes=None,
        node_type_id=None,
        custom_tags=None,
        idle_instance_autotermination_minutes=None,
        enable_elastic_disk=None,
        disk_spec=None,
        preloaded_spark_versions=None,
        headers=None,
        preloaded_docker_images=None,
        azure_attributes=None,
        gcp_attributes=None,
    ):
        _data = {}
        if instance_pool_name is not None:
            _data['instance_pool_name'] = instance_pool_name
        if min_idle_instances is not None:
            _data['min_idle_instances'] = min_idle_instances
        if max_capacity is not None:
            _data['max_capacity'] = max_capacity
        if aws_attributes is not None:
            _data['aws_attributes'] = aws_attributes
            if not isinstance(aws_attributes, dict):
                raise TypeError(
                    'Expected databricks.InstancePoolAwsAttributes() or dict for field aws_attributes'
                )
        if node_type_id is not None:
            _data['node_type_id'] = node_type_id
        if custom_tags is not None:
            _data['custom_tags'] = custom_tags
        if idle_instance_autotermination_minutes is not None:
            _data['idle_instance_autotermination_minutes'] = idle_instance_autotermination_minutes
        if enable_elastic_disk is not None:
            _data['enable_elastic_disk'] = enable_elastic_disk
        if disk_spec is not None:
            _data['disk_spec'] = disk_spec
            if not isinstance(disk_spec, dict):
                raise TypeError('Expected databricks.DiskSpec() or dict for field disk_spec')
        if preloaded_spark_versions is not None:
            _data['preloaded_spark_versions'] = preloaded_spark_versions
        if preloaded_docker_images is not None:
            _data['preloaded_docker_images'] = preloaded_docker_images
        if azure_attributes is not None:
            _data['azure_attributes'] = azure_attributes
            if not isinstance(azure_attributes, dict):
                raise TypeError(
                    'Expected databricks.InstancePoolAzureAttributes() or dict for field azure_attributes'
                )
        if gcp_attributes is not None:
            _data['gcp_attributes'] = gcp_attributes
            if not isinstance(gcp_attributes, dict):
                raise TypeError(
                    'Expected databricks.InstancePoolGcpAttributes() or dict for field gcp_attributes'
                )
        return self.client.perform_query(
            'POST', '/instance-pools/create', data=_data, headers=headers
        )

    def delete_instance_pool(self, instance_pool_id=None, headers=None):
        _data = {}
        if instance_pool_id is not None:
            _data['instance_pool_id'] = instance_pool_id
        return self.client.perform_query(
            'POST', '/instance-pools/delete', data=_data, headers=headers
        )

    def edit_instance_pool(
        self,
        instance_pool_id,
        instance_pool_name=None,
        min_idle_instances=None,
        max_capacity=None,
        aws_attributes=None,
        node_type_id=None,
        custom_tags=None,
        idle_instance_autotermination_minutes=None,
        enable_elastic_disk=None,
        disk_spec=None,
        preloaded_spark_versions=None,
        headers=None,
        preloaded_docker_images=None,
        azure_attributes=None,
        gcp_attributes=None,
    ):
        _data = {}
        if instance_pool_id is not None:
            _data['instance_pool_id'] = instance_pool_id
        if instance_pool_name is not None:
            _data['instance_pool_name'] = instance_pool_name
        if min_idle_instances is not None:
            _data['min_idle_instances'] = min_idle_instances
        if max_capacity is not None:
            _data['max_capacity'] = max_capacity
        if aws_attributes is not None:
            _data['aws_attributes'] = aws_attributes
            if not isinstance(aws_attributes, dict):
                raise TypeError(
                    'Expected databricks.InstancePoolAwsAttributes() or dict for field aws_attributes'
                )
        if node_type_id is not None:
            _data['node_type_id'] = node_type_id
        if custom_tags is not None:
            _data['custom_tags'] = custom_tags
        if idle_instance_autotermination_minutes is not None:
            _data['idle_instance_autotermination_minutes'] = idle_instance_autotermination_minutes
        if enable_elastic_disk is not None:
            _data['enable_elastic_disk'] = enable_elastic_disk
        if disk_spec is not None:
            _data['disk_spec'] = disk_spec
            if not isinstance(disk_spec, dict):
                raise TypeError('Expected databricks.DiskSpec() or dict for field disk_spec')
        if preloaded_spark_versions is not None:
            _data['preloaded_spark_versions'] = preloaded_spark_versions
        if preloaded_docker_images is not None:
            _data['preloaded_docker_images'] = preloaded_docker_images
        if azure_attributes is not None:
            _data['azure_attributes'] = azure_attributes
            if not isinstance(azure_attributes, dict):
                raise TypeError(
                    'Expected databricks.InstancePoolAzureAttributes() or dict for field azure_attributes'
                )
        if gcp_attributes is not None:
            _data['gcp_attributes'] = gcp_attributes
            if not isinstance(gcp_attributes, dict):
                raise TypeError(
                    'Expected databricks.InstancePoolGcpAttributes() or dict for field gcp_attributes'
                )
        return self.client.perform_query(
            'POST', '/instance-pools/edit', data=_data, headers=headers
        )

    def get_instance_pool(self, instance_pool_id=None, headers=None):
        _data = {}
        if instance_pool_id is not None:
            _data['instance_pool_id'] = instance_pool_id
        return self.client.perform_query('GET', '/instance-pools/get', data=_data, headers=headers)

    def list_instance_pools(self, headers=None):
        _data = {}
        return self.client.perform_query('GET', '/instance-pools/list', data=_data, headers=headers)


class DeltaPipelinesService(object):
    def __init__(self, client):
        self.client = client

    def create(
        self,
        id=None,
        name=None,
        storage=None,
        configuration=None,
        clusters=None,
        libraries=None,
        trigger=None,
        filters=None,
        target=None,
        continuous=None,
        development=None,
        allow_duplicate_names=None,
        headers=None,
    ):
        _data = {}
        if id is not None:
            _data['id'] = id
        if name is not None:
            _data['name'] = name
        if storage is not None:
            _data['storage'] = storage
        if configuration is not None:
            _data['configuration'] = configuration
        if clusters is not None:
            _data['clusters'] = clusters
        if libraries is not None:
            _data['libraries'] = libraries
        if trigger is not None:
            _data['trigger'] = trigger
            if not isinstance(trigger, dict):
                raise TypeError('Expected databricks.PipelineTrigger() or dict for field trigger')
        if filters is not None:
            _data['filters'] = filters
            if not isinstance(filters, dict):
                raise TypeError('Expected databricks.Filters() or dict for field filters')
        if target is not None:
            _data['target'] = target
        if continuous is not None:
            _data['continuous'] = continuous
        if development is not None:
            _data['development'] = development
        if allow_duplicate_names is not None:
            _data['allow_duplicate_names'] = allow_duplicate_names
        return self.client.perform_query('POST', '/pipelines', data=_data, headers=headers)

    def deploy(
        self,
        pipeline_id=None,
        id=None,
        name=None,
        storage=None,
        configuration=None,
        clusters=None,
        libraries=None,
        trigger=None,
        filters=None,
        target=None,
        continuous=None,
        development=None,
        allow_duplicate_names=None,
        headers=None,
    ):
        _data = {}
        if id is not None:
            _data['id'] = id
        if name is not None:
            _data['name'] = name
        if storage is not None:
            _data['storage'] = storage
        if configuration is not None:
            _data['configuration'] = configuration
        if clusters is not None:
            _data['clusters'] = clusters
        if libraries is not None:
            _data['libraries'] = libraries
        if trigger is not None:
            _data['trigger'] = trigger
            if not isinstance(trigger, dict):
                raise TypeError('Expected databricks.PipelineTrigger() or dict for field trigger')
        if filters is not None:
            _data['filters'] = filters
            if not isinstance(filters, dict):
                raise TypeError('Expected databricks.Filters() or dict for field filters')
        if target is not None:
            _data['target'] = target
        if continuous is not None:
            _data['continuous'] = continuous
        if development is not None:
            _data['development'] = development
        if allow_duplicate_names is not None:
            _data['allow_duplicate_names'] = allow_duplicate_names
        return self.client.perform_query(
            'PUT',
            '/pipelines/{pipeline_id}'.format(pipeline_id=pipeline_id),
            data=_data,
            headers=headers,
        )

    def delete(self, pipeline_id=None, headers=None):
        _data = {}

        return self.client.perform_query(
            'DELETE',
            '/pipelines/{pipeline_id}'.format(pipeline_id=pipeline_id),
            data=_data,
            headers=headers,
        )

    def get(self, pipeline_id=None, headers=None):
        _data = {}

        return self.client.perform_query(
            'GET',
            '/pipelines/{pipeline_id}'.format(pipeline_id=pipeline_id),
            data=_data,
            headers=headers,
        )

    def list(self, pagination=None, headers=None):
        _data = {}
        if pagination is not None:
            _data['pagination'] = pagination
            if not isinstance(pagination, dict):
                raise TypeError('Expected databricks.Pagination() or dict for field pagination')
        return self.client.perform_query('GET', '/pipelines', data=_data, headers=headers)

    def reset(self, pipeline_id=None, headers=None):
        _data = {}

        return self.client.perform_query(
            'POST',
            '/pipelines/{pipeline_id}/reset'.format(pipeline_id=pipeline_id),
            data=_data,
            headers=headers,
        )

    def run(self, pipeline_id=None, headers=None):
        _data = {}

        return self.client.perform_query(
            'POST',
            '/pipelines/{pipeline_id}/run'.format(pipeline_id=pipeline_id),
            data=_data,
            headers=headers,
        )

    def start_update(self, pipeline_id=None, full_refresh=None, headers=None):
        _data = {}
        if full_refresh:
            _data['full_refresh'] = full_refresh
        _data['cause'] = 'USER_ACTION'
        return self.client.perform_query(
            'POST',
            '/pipelines/{pipeline_id}/updates'.format(pipeline_id=pipeline_id),
            data=_data,
            headers=headers,
        )

    def stop(self, pipeline_id=None, headers=None):
        _data = {}

        return self.client.perform_query(
            'POST',
            '/pipelines/{pipeline_id}/stop'.format(pipeline_id=pipeline_id),
            data=_data,
            headers=headers,
        )


class ReposService(object):
    __git_providers__ = {
        "github.com": "gitHub",
        "dev.azure.com": "azureDevOpsServices",
        "gitlab.com": "gitLab",
        "bitbucket.org": "bitbucketCloud",
    }
    __aws_code_commit_regexp__ = re.compile(r"^git-codecommit\.[^.]+\.amazonaws.com$")

    def __init__(self, client):
        self.client = client

    @staticmethod
    def detect_repo_provider(url):
        provider = None
        try:
            netloc = urlparse(url).netloc
            idx = netloc.rfind("@")
            if idx != -1:
                netloc = netloc[(idx + 1) :]
            provider = ReposService.__git_providers__.get(netloc.lower())
            if provider is None and ReposService.__aws_code_commit_regexp__.match(netloc):
                provider = "awsCodeCommit"
        except:
            pass
        return provider

    def list_repos(self, path_prefix=None, next_page_token=None, headers=None):
        _data = {}
        if path_prefix is not None:
            _data['path_prefix'] = path_prefix
        if next_page_token is not None:
            _data['next_page_token'] = next_page_token
        return self.client.perform_query('GET', '/repos', data=_data, headers=headers)

    def get_repo(self, id, headers=None):
        _data = {}

        return self.client.perform_query(
            'GET', '/repos/{id}'.format(id=id), data=_data, headers=headers
        )

    def update_repo(self, id, branch=None, tag=None, headers=None):
        _data = {}
        if branch is not None:
            _data['branch'] = branch
        if tag is not None:
            _data['tag'] = tag
        return self.client.perform_query(
            'PATCH', '/repos/{id}'.format(id=id), data=_data, headers=headers
        )

    def create_repo(self, url, provider, path=None, headers=None):
        _data = {}
        if url is not None:
            _data['url'] = url
        if provider is None or provider.strip() == "":
            provider = self.detect_repo_provider(url)
        if provider is not None:
            _data['provider'] = provider
        else:
            raise ValueError(
                "The Git provider parameter wasn't specified and we can't detect it "
                "from URL. Please pass 'provider' option"
            )
        if path is not None:
            _data['path'] = path
        return self.client.perform_query('POST', '/repos', data=_data, headers=headers)

    def delete_repo(self, id, headers=None):
        _data = {}

        return self.client.perform_query(
            'DELETE', '/repos/{id}'.format(id=id), data=_data, headers=headers
        )



print("import: 'databricks_cli'")
import databricks_cli

print("import: 'databricks_cli.cluster_policies'")
import databricks_cli.cluster_policies

print("import: 'databricks_cli.clusters'")
import databricks_cli.clusters

print("import: 'databricks_cli.configure'")
import databricks_cli.configure

print("import: 'databricks_cli.dbfs'")
import databricks_cli.dbfs

print("import: 'databricks_cli.groups'")
import databricks_cli.groups

print("import: 'databricks_cli.instance_pools'")
import databricks_cli.instance_pools

print("import: 'databricks_cli.jobs'")
import databricks_cli.jobs

print("import: 'databricks_cli.libraries'")
import databricks_cli.libraries

print("import: 'databricks_cli.oauth'")
import databricks_cli.oauth

print("import: 'databricks_cli.pipelines'")
import databricks_cli.pipelines

print("import: 'databricks_cli.repos'")
import databricks_cli.repos

print("import: 'databricks_cli.runs'")
import databricks_cli.runs

print("import: 'databricks_cli.sdk'")
import databricks_cli.sdk

print("import: 'databricks_cli.secrets'")
import databricks_cli.secrets

print("import: 'databricks_cli.stack'")
import databricks_cli.stack

print("import: 'databricks_cli.tokens'")
import databricks_cli.tokens

print("import: 'databricks_cli.unity_catalog'")
import databricks_cli.unity_catalog

print("import: 'databricks_cli.workspace'")
import databricks_cli.workspace




# Databricks CLI
# Copyright 2017 Databricks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"), except
# that the use of services to which certain application programming
# interfaces (each, an "API") connect requires that the user first obtain
# a license for the use of the APIs from Databricks, Inc. ("Databricks"),
# by creating an account at www.databricks.com and agreeing to either (a)
# the Community Edition Terms of Service, (b) the Databricks Terms of
# Service, or (c) another written agreement between Licensee and Databricks
# for the use of the APIs.
#
# You may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from json import loads as json_loads

import click
from tabulate import tabulate

from databricks_cli.click_types import OutputClickType, JsonClickType, JobIdClickType
from databricks_cli.jobs.api import JobsApi
from databricks_cli.utils import eat_exceptions, CONTEXT_SETTINGS, pretty_format, json_cli_base, \
    truncate_string

from databricks_cli.configure.config import provide_api_client, profile_option, \
    get_profile_from_context, debug_option, get_config, api_version_option
from databricks_cli.configure.provider import DatabricksConfig, update_and_persist_config, \
    ProfileConfigProvider
from databricks_cli.version import print_version_callback, version as cli_version


@click.command(context_settings=CONTEXT_SETTINGS)
@click.option('--json-file', default=None, type=click.Path(),
              help='File containing JSON request to POST to /api/2.*/jobs/create.')
@click.option('--json', default=None, type=JsonClickType(),
              help=JsonClickType.help('/api/2.*/jobs/create'))
@api_version_option
@debug_option
@profile_option
@eat_exceptions
@provide_api_client
def create_cli(api_client, json_file, json, version):
    """
    Creates a job.

    The specification for the json option can be found
    https://docs.databricks.com/api/latest/jobs.html#create
    """
    check_version(api_client, version)
    json_cli_base(json_file, json,
                  lambda json: JobsApi(api_client).create_job(json, version=version))


@click.command(context_settings=CONTEXT_SETTINGS)
@click.option('--job-id', default=None, type=JobIdClickType(), help=JobIdClickType.help)
@click.option('--json-file', default=None, type=click.Path(),
              help='File containing partial JSON request to POST to /api/2.*/jobs/reset. '
                   'For more, read full help message.')
@click.option('--json', default=None, type=JsonClickType(),
              help='Partial JSON string to POST to /api/2.*/jobs/reset. '
                   'For more, read full help message.')
@api_version_option
@debug_option
@profile_option
@eat_exceptions
@provide_api_client
def reset_cli(api_client, json_file, json, job_id, version):
    """
    Resets (edits) the definition of a job.

    The specification for the json option can be found
    https://docs.databricks.com/api/latest/jobs.html#jobsjobsettings
    """
    check_version(api_client, version)
    if not bool(json_file) ^ bool(json):
        raise RuntimeError('Either --json-file or --json should be provided')
    if json_file:
        with open(json_file, 'r') as f:
            json = f.read()
    deser_json = json_loads(json)
    # If the payload is defined using the API definition rather than the CLI
    # extract the settings data.
    new_settings = deser_json['new_settings'] if (
        'new_settings' in deser_json) else deser_json

    # If job id is not defined in the call we fall back to checking
    # the JSON for a job_id property
    if job_id is None:
        if 'job_id' in deser_json:
            job_id = deser_json['job_id']
        else:
            raise RuntimeError(
                'Either --job-id or a root-level json key "job_id" should be provided')

    request_body = {
        'job_id': job_id,
        'new_settings': new_settings
    }

    JobsApi(api_client).reset_job(request_body, version=version)


def _jobs_to_table(jobs_json):
    ret = []
    for j in jobs_json['jobs']:
        ret.append((j['job_id'], truncate_string(j['settings']['name'])))
    return sorted(ret, key=lambda t: t[1].lower())


@click.command(context_settings=CONTEXT_SETTINGS,
               short_help='Lists the jobs in the Databricks Job Service.')
@click.option('--output', default=None, help=OutputClickType.help, type=OutputClickType())
@click.option('--type', 'job_type', default=None, help='The type of job to list', type=str)
@click.option('--expand-tasks', is_flag=True,
              help='Expands the tasks array (only available in API 2.1).')
@click.option('--offset', default=None, type=int,
              help='The offset to use when listing jobs (only available in API 2.1).')
@click.option('--limit', default=None, type=int,
              help='The maximum number of jobs to fetch in a single call ' +
                   '(only available in API 2.1).')
@click.option('--all', '_all', is_flag=True,
              help='Lists all jobs by executing sequential calls to the API ' +
                   '(only available in API 2.1).')
@click.option('--name', 'name', default=None, type=str,
              help='If provided, only returns jobs that match the supplied ' + 
                   'name (only available in API 2.1).')
@api_version_option
@debug_option
@profile_option
@eat_exceptions
@provide_api_client
def list_cli(api_client, output, job_type, version, expand_tasks, offset, limit, _all, name):
    """
    Lists the jobs in the Databricks Job Service.

    By default the output format will be a human readable table with the following fields

      - Job ID

      - Job name

    A JSON formatted output can also be requested by setting the --output parameter to "JSON"

    In table mode, the jobs are sorted by their name.
    """
    check_version(api_client, version)
    api_version = version or api_client.jobs_api_version
    using_features_only_in_21 = expand_tasks or offset or limit or _all or name
    if api_version != '2.1' and using_features_only_in_21:
        click.echo(click.style('ERROR', fg='red') + ': the options --expand-tasks, ' +
                   '--offset, --limit, --all, and --name are only available in API 2.1', err=True)
        return
    jobs_api = JobsApi(api_client)
    has_more = True
    jobs = []
    if _all:
        offset = 0
        limit = 20
    while has_more:
        jobs_json = jobs_api.list_jobs(job_type=job_type, expand_tasks=expand_tasks,
                                       offset=offset, limit=limit, version=version,
                                       name=name)
        jobs += jobs_json['jobs'] if 'jobs' in jobs_json else []
        has_more = jobs_json.get('has_more', False) and _all
        if has_more:
            offset = offset + \
                (len(jobs_json['jobs']) if 'jobs' in jobs_json else 20)

    out = {'jobs': jobs}
    if OutputClickType.is_json(output):
        click.echo(pretty_format(out))
    else:
        click.echo(tabulate(_jobs_to_table(out),
                   tablefmt='plain', disable_numparse=True))


@click.command(context_settings=CONTEXT_SETTINGS,
               short_help='Deletes the specified job.')
@click.option('--job-id', required=True, type=JobIdClickType(), help=JobIdClickType.help)
@api_version_option
@debug_option
@profile_option
@eat_exceptions
@provide_api_client
def delete_cli(api_client, job_id, version):
    """
    Deletes the specified job.
    """
    check_version(api_client, version)
    JobsApi(api_client).delete_job(job_id, version=version)


@click.command(context_settings=CONTEXT_SETTINGS)
@click.option('--job-id', required=True, type=JobIdClickType(), help=JobIdClickType.help)
@api_version_option
@debug_option
@profile_option
@eat_exceptions
@provide_api_client
def get_cli(api_client, job_id, version):
    """
    Describes the metadata for a job.
    """
    check_version(api_client, version)
    click.echo(pretty_format(
        JobsApi(api_client).get_job(job_id, version=version)))


@click.command(context_settings=CONTEXT_SETTINGS)
@click.option('--job-id', required=True, type=JobIdClickType(), help=JobIdClickType.help)
@click.option('--jar-params', default=None, type=JsonClickType(),
              help='JSON string specifying an array of parameters. i.e. ["param1", "param2"]')
@click.option('--notebook-params', default=None, type=JsonClickType(),
              help='JSON string specifying a map of key-value pairs. '
                   'i.e. {"name": "john doe", "age": 35}')
@click.option('--python-params', default=None, type=JsonClickType(),
              help='JSON string specifying an array of parameters. i.e. ["param1", "param2"]')
@click.option('--python-named-params', default=None, type=JsonClickType(),
              help='JSON string specifying a map of key-value pairs. '
                   'i.e. {"name": "john doe", "age": 35}')
@click.option('--spark-submit-params', default=None, type=JsonClickType(),
              help='JSON string specifying an array of parameters. i.e. '
                   '["--class", "org.apache.spark.examples.SparkPi"]')
@click.option('--idempotency-token', default=None,
              help='If an active run with the provided token already exists, ' +
              'the request does not create a new run, ' +
              'but returns the ID of the existing run instead.')
@api_version_option
@debug_option
@profile_option
@eat_exceptions
@provide_api_client
def run_now_cli(api_client, job_id, jar_params, notebook_params, python_params,
                python_named_params, spark_submit_params, idempotency_token, version):
    """
    Runs a job with optional per-run parameters.

    Parameter options are specified in json and the format is documented in
    https://docs.databricks.com/api/latest/jobs.html#jobsrunnow.
    """
    check_version(api_client, version)
    jar_params_json = json_loads(jar_params) if jar_params else None
    notebook_params_json = json_loads(
        notebook_params) if notebook_params else None
    python_params = json_loads(python_params) if python_params else None
    python_named_params = json_loads(
        python_named_params) if python_named_params else None
    spark_submit_params = json_loads(
        spark_submit_params) if spark_submit_params else None
    res = JobsApi(api_client).run_now(
        job_id, jar_params_json, notebook_params_json, python_params,
        spark_submit_params, python_named_params, idempotency_token, version=version)
    click.echo(pretty_format(res))


@click.command(context_settings=CONTEXT_SETTINGS)
@api_version_option
@debug_option
@profile_option
def configure(version):
    profile = get_profile_from_context()
    config = ProfileConfigProvider(
        profile).get_config() if profile else get_config()
    new_config = config or DatabricksConfig.empty()
    new_config.jobs_api_version = version
    update_and_persist_config(profile, new_config)


@click.group(context_settings=CONTEXT_SETTINGS,
             short_help='Utility to interact with jobs.')
@click.option('--version', '-v', is_flag=True, callback=print_version_callback,
              expose_value=False, is_eager=True, help=cli_version)
@debug_option
@profile_option
@eat_exceptions
def jobs_group():  # pragma: no cover
    """
    Utility to interact with jobs.

    This is a wrapper around the jobs API (https://docs.databricks.com/api/latest/jobs.html).
    Job runs are handled by ``databricks runs``.
    """
    pass


jobs_group.add_command(create_cli, name='create')
jobs_group.add_command(list_cli, name='list')
jobs_group.add_command(delete_cli, name='delete')
jobs_group.add_command(get_cli, name='get')
jobs_group.add_command(reset_cli, name='reset')
jobs_group.add_command(run_now_cli, name='run-now')
jobs_group.add_command(configure, name='configure')


def check_version(api_client, version):
    if version is not None:
        # If the user explicitly passed --version=2.x for this invocation it means
        # they really really want that version, let's not show any warnings
        return

    if api_client.jobs_api_version == '2.1':
        # If the user is globally configured to use 2.1 we don't show the warning
        return

    click.echo(click.style('WARN', fg='yellow') + ': Your CLI is configured ' +
               'to use Jobs API 2.0. In order to use the latest Jobs features ' +
               'please upgrade to 2.1: \'databricks jobs configure --version=2.1\'. ' +
               'Future versions of this CLI will default to the new Jobs API. ' +
               'Learn more at https://docs.databricks.com/dev-tools/cli/jobs-cli.html',
               err=True
               )



# Databricks CLI
# Copyright 2017 Databricks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"), except
# that the use of services to which certain application programming
# interfaces (each, an "API") connect requires that the user first obtain
# a license for the use of the APIs from Databricks, Inc. ("Databricks"),
# by creating an account at www.databricks.com and agreeing to either (a)
# the Community Edition Terms of Service, (b) the Databricks Terms of
# Service, or (c) another written agreement between Licensee and Databricks
# for the use of the APIs.
#
# You may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import click

from databricks_cli.configure.config import provide_api_client, profile_option, debug_option
from databricks_cli.tokens.api import TokensApi
from databricks_cli.utils import CONTEXT_SETTINGS, eat_exceptions, pretty_format
from databricks_cli.version import print_version_callback, version


TOKEN_LIFETIME_SEC = 60 * 60 * 24 * 90


@click.command(context_settings=CONTEXT_SETTINGS,
               short_help='Create a token.')
@click.option('--lifetime-seconds', required=False, default=TOKEN_LIFETIME_SEC,
              help="Number of seconds for the token to live for. The default is %d seconds or "
                   "%d days." % (TOKEN_LIFETIME_SEC, TOKEN_LIFETIME_SEC / 60 / 60 / 24))
@click.option('--comment', required=True,
              help="String describing what the token is for.")
@debug_option
@profile_option
@eat_exceptions  # noqa
@provide_api_client
def create_token_cli(api_client, lifetime_seconds, comment):
    """
    Create and return a token.
    This call returns the error QUOTA_EXCEEDED if the caller exceeds the token quota, which is 600.
    """
    content = TokensApi(api_client).create(lifetime_seconds, comment)
    click.echo(pretty_format(content))


@click.command(context_settings=CONTEXT_SETTINGS,
               short_help='List tokens for the calling user.')
@debug_option
@profile_option
@eat_exceptions  # noqa
@provide_api_client
def list_cli(api_client):
    """
    List all the valid tokens for a user-workspace pair.
    """
    content = TokensApi(api_client).list()
    click.echo(pretty_format(content))


@click.command(context_settings=CONTEXT_SETTINGS,
               short_help='Revoke an access token.')
@click.option('--token-id', required=True,
              help="Token id to revoke")
@debug_option
@profile_option
@eat_exceptions  # noqa
@provide_api_client
def revoke_cli(api_client, token_id):
    """
    Revoke an access token.

    This call returns the error RESOURCE_DOES_NOT_EXIST
    if a token with the specified ID is not valid.
    """
    content = TokensApi(api_client).revoke(token_id)
    click.echo(pretty_format(content))


@click.group(context_settings=CONTEXT_SETTINGS,
             short_help="Utility to interact with Databricks tokens.")
@click.option("--version", "-v", is_flag=True, callback=print_version_callback,
              expose_value=False, is_eager=True, help=version)
@debug_option
@profile_option
@eat_exceptions
def tokens_group():  # pragma: no cover
    """Utility to interact with Databricks tokens."""
    pass


tokens_group.add_command(create_token_cli, name='create')
tokens_group.add_command(list_cli, name='list')
tokens_group.add_command(revoke_cli, name='revoke')



# Databricks CLI
# Copyright 2017 Databricks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"), except
# that the use of services to which certain application programming
# interfaces (each, an "API") connect requires that the user first obtain
# a license for the use of the APIs from Databricks, Inc. ("Databricks"),
# by creating an account at www.databricks.com and agreeing to either (a)
# the Community Edition Terms of Service, (b) the Databricks Terms of
# Service, or (c) another written agreement between Licensee and Databricks
# for the use of the APIs.
#
# You may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from base64 import b64encode, b64decode

import click

from databricks_cli.dbfs.exceptions import LocalFileExistsException
from databricks_cli.sdk import WorkspaceService
from databricks_cli.workspace.types import WorkspaceFormat, WorkspaceLanguage

DIRECTORY = 'DIRECTORY'
NOTEBOOK = 'NOTEBOOK'
LIBRARY = 'LIBRARY'
REPO = 'REPO'


class WorkspaceFileInfo(object):
    def __init__(self, path, object_type, object_id, language=None, **kwargs): # noqa
        self.path = path
        self.object_type = object_type
        self.language = language
        self.object_id = object_id

    def to_row(self, is_long_form, is_absolute, with_object_id=False):
        path = self.path if is_absolute else self.basename
        if self.is_dir:
            stylized_path = click.style(path, 'cyan')
        elif self.is_library:
            stylized_path = click.style(path, 'green')
        else:
            stylized_path = path

        result = [stylized_path]

        if is_long_form:
            result = [self.object_type, stylized_path, self.language]

        if with_object_id:
            result.append(self.object_id)

        return result

    @property
    def is_dir(self):
        return self.object_type == DIRECTORY

    @property
    def is_notebook(self):
        return self.object_type == NOTEBOOK

    @property
    def is_library(self):
        return self.object_type == LIBRARY

    @property
    def basename(self):
        return os.path.basename(self.path)

    @classmethod
    def from_json(cls, deserialized_json):
        return cls(**deserialized_json)


class WorkspaceApi(object):
    def __init__(self, api_client):
        self.client = WorkspaceService(api_client)

    def get_status(self, workspace_path, headers=None):
        return WorkspaceFileInfo.from_json(self.client.get_status(workspace_path, headers=headers))

    def list_objects(self, workspace_path, headers=None):
        response = self.client.list(workspace_path, headers=headers)
        # This case is necessary when we list an empty dir in the workspace.
        # TODO(andrewmchen): We should make our API respond with a json with 'objects' field even
        # in this case.
        if 'objects' not in response:
            return []
        objects = response['objects']
        return [WorkspaceFileInfo.from_json(f) for f in objects]

    def mkdirs(self, workspace_path, headers=None):
        self.client.mkdirs(workspace_path, headers=headers)

    def import_workspace(self, source_path, target_path, language, fmt, is_overwrite, headers=None):
        with open(source_path, 'rb') as f:
            # import_workspace must take content that is typed str.
            content = b64encode(f.read()).decode()
            self.client.import_workspace(
                target_path,
                fmt,
                language,
                content,
                is_overwrite,
                headers=headers)

    def export_workspace(self, source_path, target_path, fmt, is_overwrite, headers=None):
        """
        Faithfully exports the source_path to the target_path. Does not
        attempt to do any munging of the target_path if it is a directory.
        """
        if os.path.exists(target_path) and not is_overwrite:
            raise LocalFileExistsException('Target {} already exists.'.format(target_path))
        output = self.client.export_workspace(source_path, fmt, headers=headers)
        content = output['content']
        # Will overwrite target_path.
        with open(target_path, 'wb') as f:
            decoded = b64decode(content)
            f.write(decoded)

    def delete(self, workspace_path, is_recursive, headers=None):
        self.client.delete(workspace_path, is_recursive, headers=headers)

    def import_workspace_dir(self, source_path, target_path, overwrite, exclude_hidden_files,
                             headers=None):
        # pylint: disable=too-many-locals
        filenames = os.listdir(source_path)
        if exclude_hidden_files:
            # for now, just exclude hidden files or directories based on starting '.'
            filenames = [f for f in filenames if not f.startswith('.')]
        self.mkdirs(target_path, headers=headers)
        for filename in filenames:
            cur_src = os.path.join(source_path, filename)
            # don't use os.path.join here since it will set \ on Windows
            cur_dst = target_path.rstrip('/') + '/' + filename
            if os.path.isdir(cur_src):
                self.import_workspace_dir(cur_src, cur_dst, overwrite, exclude_hidden_files,
                                          headers=headers)
            elif os.path.isfile(cur_src):
                ext = WorkspaceLanguage.get_extension(cur_src)
                if ext != '':
                    cur_dst = cur_dst[:-len(ext)]
                    (language, file_format) = WorkspaceLanguage.to_language_and_format(cur_src)
                    self.import_workspace(cur_src, cur_dst, language, file_format, overwrite,
                                          headers=headers)
                    click.echo('{} -> {}'.format(cur_src, cur_dst))
                else:
                    extensions = ', '.join(WorkspaceLanguage.EXTENSIONS)
                    click.echo(('{} does not have a valid extension of {}. Skip this file and ' +
                                'continue.').format(cur_src, extensions))

    def export_workspace_dir(self, source_path, target_path, overwrite, headers=None):
        if os.path.isfile(target_path):
            click.echo('{} exists as a file. Skipping this subtree {}'
                       .format(target_path, source_path))
            return
        if not os.path.isdir(target_path):
            os.makedirs(target_path)
        for obj in self.list_objects(source_path, headers=headers):
            cur_src = obj.path
            cur_dst = os.path.join(target_path, obj.basename)
            if obj.is_dir:
                self.export_workspace_dir(cur_src, cur_dst, overwrite, headers=headers)
            elif obj.is_notebook:
                cur_dst = cur_dst + WorkspaceLanguage.to_extension(obj.language)
                try:
                    self.export_workspace(cur_src, cur_dst, WorkspaceFormat.SOURCE, overwrite,
                                          headers=headers)
                    click.echo('{} -> {}'.format(cur_src, cur_dst))
                except LocalFileExistsException:
                    click.echo('{} already exists locally as {}. Skip.'.format(cur_src, cur_dst))
            else:
                click.echo('{} is neither a dir or a notebook. Skip.'.format(cur_src))



# Databricks CLI
# Copyright 2021 Databricks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"), except
# that the use of services to which certain application programming
# interfaces (each, an "API") connect requires that the user first obtain
# a license for the use of the APIs from Databricks, Inc. ("Databricks"),
# by creating an account at www.databricks.com and agreeing to either (a)
# the Community Edition Terms of Service, (b) the Databricks Terms of
# Service, or (c) another written agreement between Licensee and Databricks
# for the use of the APIs.
#
# You may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import click

from databricks_cli.utils import CONTEXT_SETTINGS
from databricks_cli.version import print_version_callback, version

from databricks_cli.unity_catalog.metastore_cli import register_metastore_commands
from databricks_cli.unity_catalog.catalog_cli import register_catalog_commands
from databricks_cli.unity_catalog.schema_cli import register_schema_commands
from databricks_cli.unity_catalog.table_cli import register_table_commands
from databricks_cli.unity_catalog.ext_loc_cli import register_ext_loc_commands
from databricks_cli.unity_catalog.cred_cli import register_cred_commands
from databricks_cli.unity_catalog.delta_sharing_cli import register_delta_sharing_commands
from databricks_cli.unity_catalog.perms_cli import register_perms_commands
from databricks_cli.unity_catalog.lineage_cli import register_lineage_commands


@click.group(context_settings=CONTEXT_SETTINGS)
@click.option('--version', '-v', is_flag=True, callback=print_version_callback,
              expose_value=False, is_eager=True, help=version)
def unity_catalog_group():  # pragma: no cover
    """
    Utility to interact with Databricks Unity Catalog.
    """
    pass


register_metastore_commands(unity_catalog_group)
register_ext_loc_commands(unity_catalog_group)
register_cred_commands(unity_catalog_group)
register_catalog_commands(unity_catalog_group)
register_schema_commands(unity_catalog_group)
register_table_commands(unity_catalog_group)
register_delta_sharing_commands(unity_catalog_group)
register_perms_commands(unity_catalog_group)
register_lineage_commands(unity_catalog_group)



# Databricks CLI
# Copyright 2017 Databricks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"), except
# that the use of services to which certain application programming
# interfaces (each, an "API") connect requires that the user first obtain
# a license for the use of the APIs from Databricks, Inc. ("Databricks"),
# by creating an account at www.databricks.com and agreeing to either (a)
# the Community Edition Terms of Service, (b) the Databricks Terms of
# Service, or (c) another written agreement between Licensee and Databricks
# for the use of the APIs.
#
# You may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from json import loads as json_loads

import click
from tabulate import tabulate

from databricks_cli.click_types import OutputClickType, JsonClickType, ClusterPolicyIdClickType
from databricks_cli.cluster_policies.api import ClusterPolicyApi
from databricks_cli.utils import eat_exceptions, CONTEXT_SETTINGS, pretty_format, json_cli_base, \
    truncate_string
from databricks_cli.configure.config import provide_api_client, profile_option, debug_option
from databricks_cli.version import print_version_callback, version


@click.command(context_settings=CONTEXT_SETTINGS)
@click.option('--json-file', default=None, type=click.Path(),
              help='File containing JSON request to POST to /api/2.0/policies/clusters/create.')
@click.option('--json', default=None, type=JsonClickType(),
              help=JsonClickType.help('/api/2.0/policies/clusters/create'))
@debug_option
@profile_option
@eat_exceptions
@provide_api_client
def create_cli(api_client, json_file, json):
    """
    Creates a Databricks cluster Policy.

    The specification for the request json can be found at
    https://docs.databricks.com/dev-tools/api/latest/policies.html#create
    """
    json_cli_base(json_file, json,
                  lambda json: ClusterPolicyApi(api_client).create_cluster_policy(json))


@click.command(context_settings=CONTEXT_SETTINGS)
@click.option('--json-file', default=None, type=click.Path(),
              help='File containing JSON request to POST to /api/2.0/policies/clusters/edit.')
@click.option('--json', default=None, type=JsonClickType(),
              help=JsonClickType.help('/api/2.0/policies/clusters/edit'))
@debug_option
@profile_option
@eat_exceptions
@provide_api_client
def edit_cli(api_client, json_file, json):
    """
    Edits a Databricks cluster Policy.

    The specification for the request json can be found at
    https://docs.databricks.com/dev-tools/api/latest/policies.html#edit
    """
    if not bool(json_file) ^ bool(json):
        raise RuntimeError('Either --json-file or --json should be provided')
    if json_file:
        with open(json_file, 'r') as f:
            json = f.read()
    deser_json = json_loads(json)
    ClusterPolicyApi(api_client).edit_cluster_policy(deser_json)


@click.command(context_settings=CONTEXT_SETTINGS)
@click.option('--policy-id', required=True, type=ClusterPolicyIdClickType(),
              help=ClusterPolicyIdClickType.help)
@debug_option
@profile_option
@eat_exceptions
@provide_api_client
def delete_cli(api_client, policy_id):
    """
    Removes a Databricks cluster policy given its ID.

    Use ``databricks cluster_olicies get --policy-id POLICY_ID`` to check termination states.
    """
    ClusterPolicyApi(api_client).delete_cluster_policy(policy_id)


@click.command(context_settings=CONTEXT_SETTINGS)
@click.option('--policy-id', required=True, type=ClusterPolicyIdClickType(),
              help=ClusterPolicyIdClickType.help)
@debug_option
@profile_option
@eat_exceptions
@provide_api_client
def get_cli(api_client, policy_id):
    """
    Retrieves metadata about a cluster policy.
    """
    click.echo(pretty_format(ClusterPolicyApi(api_client).get_cluster_policy(policy_id)))


def _cluster_policies_to_table(policies_json):
    ret = []
    for c in policies_json.get('policies', []):
        ret.append((c['policy_id'], truncate_string(c['name']), c['definition']))
    return ret


@click.command(context_settings=CONTEXT_SETTINGS,
               short_help='Lists Cluster Policies.')
@click.option('--output', default=None, help=OutputClickType.help, type=OutputClickType())
@debug_option
@profile_option
@eat_exceptions
@provide_api_client
def list_cli(api_client, output):
    """
    Lists cluster olicies.

    Returns information about all currently cluster olicies.

    By default the output format will be a human readable table with the following fields

      - Policy ID

      - Policy Name

      - Policy Definition
    """
    policies_json = ClusterPolicyApi(api_client).list_cluster_policies()

    if OutputClickType.is_json(output):
        click.echo(pretty_format(policies_json))
    else:
        click.echo(tabulate(_cluster_policies_to_table(policies_json), tablefmt='plain'))


@click.group(context_settings=CONTEXT_SETTINGS,
             short_help='Utility to interact with Databricks cluster policies.')
@click.option('--version', '-v', is_flag=True, callback=print_version_callback,
              expose_value=False, is_eager=True, help=version)
@debug_option
@profile_option
@eat_exceptions
def cluster_policies_group():  # pragma: no cover
    """
    Utility to interact with Databricks cluster policies.
    """
    pass


cluster_policies_group.add_command(create_cli, name='create')
cluster_policies_group.add_command(edit_cli, name='edit')
cluster_policies_group.add_command(delete_cli, name='delete')
cluster_policies_group.add_command(get_cli, name='get')
cluster_policies_group.add_command(list_cli, name='list')



# Databricks CLI
# Copyright 2022 Databricks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"), except
# that the use of services to which certain application programming
# interfaces (each, an "API") connect requires that the user first obtain
# a license for the use of the APIs from Databricks, Inc. ("Databricks"),
# by creating an account at www.databricks.com and agreeing to either (a)
# the Community Edition Terms of Service, (b) the Databricks Terms of
# Service, or (c) another written agreement between Licensee and Databricks
# for the use of the APIs.
#
# You may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import click

from databricks_cli.click_types import JsonClickType
from databricks_cli.configure.config import provide_api_client, profile_option, debug_option
from databricks_cli.unity_catalog.api import UnityCatalogApi
from databricks_cli.unity_catalog.utils import hide, json_file_help, json_string_help, \
    mc_pretty_format
from databricks_cli.utils import eat_exceptions, CONTEXT_SETTINGS, json_cli_base


@click.command(context_settings=CONTEXT_SETTINGS,
               short_help='Create a new catalog.')
@click.option('--name', required=True, help='Name of new catalog.')
@click.option('--comment', default=None, required=False,
              help='Free-form text description.')
@click.option('--provider', default=None, required=False,
              help='Name of the Provider (for creating Delta Sharing Catalog).')
@click.option('--share', default=None, required=False,
              help='Name of the Share under the Provider to create a Delta Sharing Catalog.')
@debug_option
@profile_option
@eat_exceptions
@provide_api_client
def create_catalog_cli(api_client, name, comment, provider, share):
    """
    Create a new catalog.
    """
    catalog_json = UnityCatalogApi(api_client).create_catalog(name, comment,
                                                              provider, share)
    click.echo(mc_pretty_format(catalog_json))


@click.command(context_settings=CONTEXT_SETTINGS,
               short_help='List catalogs.')
@debug_option
@profile_option
@eat_exceptions
@provide_api_client
def list_catalogs_cli(api_client):
    """
    List catalogs.
    """
    catalogs_json = UnityCatalogApi(api_client).list_catalogs()
    click.echo(mc_pretty_format(catalogs_json))


@click.command(context_settings=CONTEXT_SETTINGS,
               short_help='Get a catalog.')
@click.option('--name', required=True,
              help='Name of the catalog to get.')
@debug_option
@profile_option
@eat_exceptions
@provide_api_client
def get_catalog_cli(api_client, name):
    """
    Get a catalog.
    """
    catalog_json = UnityCatalogApi(api_client).get_catalog(name)
    click.echo(mc_pretty_format(catalog_json))


@click.command(context_settings=CONTEXT_SETTINGS,
               short_help='Update a catalog.')
@click.option('--name', required=True,
              help='Name of the catalog to update.')
@click.option('--json-file', default=None, type=click.Path(),
              help=json_file_help(method='PATCH', path='/catalogs/{name}'))
@click.option('--json', default=None, type=JsonClickType(),
              help=json_string_help(method='PATCH', path='/catalogs/{name}'))
@debug_option
@profile_option
@eat_exceptions
@provide_api_client
def update_catalog_cli(api_client, name, json_file, json):
    """
    Update a catalog.

    The public specification for the JSON request is in development.
    """
    json_cli_base(json_file, json,
                  lambda json: UnityCatalogApi(api_client).update_catalog(name, json))


@click.command(context_settings=CONTEXT_SETTINGS,
               short_help='Delete a catalog.')
@click.option('--name', required=True,
              help='Name of the catalog to delete.')
@click.option('--purge', '-p', is_flag=True, default=False,
              help='Purge all child schemas and tables of catalog.')
@debug_option
@profile_option
@eat_exceptions
@provide_api_client
def delete_catalog_cli(api_client, name, purge):
    """
    Delete a catalog.
    """
    if purge:
        tables_response = UnityCatalogApi(api_client).list_table_summaries(name)
        tables = tables_response.get('tables', [])
        tables = filter(lambda t: t['full_name'].split('.')[1] != 'information_schema', tables)
        for t in tables:
            click.echo("Deleting table: %s" % (t['full_name']))
            UnityCatalogApi(api_client).delete_table(t['full_name'])

        schemas_response = UnityCatalogApi(api_client).list_schemas(name, None)
        schemas = schemas_response.get('schemas', [])
        schemas = filter(lambda s: s['name'] != 'information_schema', schemas)
        for s in schemas:
            click.echo("Purging schema: %s" % (s['full_name']))
            UnityCatalogApi(api_client).delete_schema(s['full_name'])

    UnityCatalogApi(api_client).delete_catalog(name)


@click.group()
def catalogs_group():  # pragma: no cover
    pass


def register_catalog_commands(cmd_group):
    # Register deprecated "verb-noun" commands for backward compatibility.
    cmd_group.add_command(hide(create_catalog_cli), name='create-catalog')
    cmd_group.add_command(hide(list_catalogs_cli), name='list-catalogs')
    cmd_group.add_command(hide(get_catalog_cli), name='get-catalog')
    cmd_group.add_command(hide(update_catalog_cli), name='update-catalog')
    cmd_group.add_command(hide(delete_catalog_cli), name='delete-catalog')

    # Register command group.
    catalogs_group.add_command(create_catalog_cli, name='create')
    catalogs_group.add_command(list_catalogs_cli, name='list')
    catalogs_group.add_command(get_catalog_cli, name='get')
    catalogs_group.add_command(update_catalog_cli, name='update')
    catalogs_group.add_command(delete_catalog_cli, name='delete')
    cmd_group.add_command(catalogs_group, name='catalogs')



# Databricks CLI
# Copyright 2022 Databricks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"), except
# that the use of services to which certain application programming
# interfaces (each, an "API") connect requires that the user first obtain
# a license for the use of the APIs from Databricks, Inc. ("Databricks"),
# by creating an account at www.databricks.com and agreeing to either (a)
# the Community Edition Terms of Service, (b) the Databricks Terms of
# Service, or (c) another written agreement between Licensee and Databricks
# for the use of the APIs.
#
# You may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy

from databricks_cli.utils import pretty_format


# Encode UTF-8 strings in JSON blobs
def mc_pretty_format(json):
    return pretty_format(json, encode_utf8=True)


def del_none(json):
    to_delete = []
    to_recurse = []
    for key, value in json.items():
        if value is None:
            to_delete.append(key)
        if value is dict:
            to_recurse.append(key)
    for key in to_delete:
        del json[key]
    for key in to_recurse:
        del_none(json[key])


def hide(cmd):
    """
    Return a copy of specified Click command instance with `hidden = True`.
    This requires Click >= v7.0.
    """
    cmd_copy = copy.copy(cmd)
    cmd_copy.hidden = True
    return cmd_copy


def json_file_help(method, path):
    path = "/api/2.0/unity-catalog" + path
    return "File containing JSON request to {} to {}.".format(method, path)


def json_string_help(method, path):
    path = "/api/2.0/unity-catalog" + path
    return "JSON string to {} to {}.".format(method, path)



# Databricks CLI
# Copyright 2022 Databricks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"), except
# that the use of services to which certain application programming
# interfaces (each, an "API") connect requires that the user first obtain
# a license for the use of the APIs from Databricks, Inc. ("Databricks"),
# by creating an account at www.databricks.com and agreeing to either (a)
# the Community Edition Terms of Service, (b) the Databricks Terms of
# Service, or (c) another written agreement between Licensee and Databricks
# for the use of the APIs.
#
# You may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import click

from databricks_cli.configure.config import provide_api_client, profile_option, debug_option
from databricks_cli.unity_catalog.api import UnityCatalogApi
from databricks_cli.unity_catalog.utils import mc_pretty_format, hide
from databricks_cli.utils import eat_exceptions, CONTEXT_SETTINGS, to_graph


@click.command(context_settings=CONTEXT_SETTINGS,
               short_help='List table lineage.')
@click.option('--table-name', required=True,
              help='Name of the table with 3L namespace')
@click.option('--level', required=False, type=int, default=1,
              help='level of lineage to retrieve')
@debug_option
@profile_option
@eat_exceptions
@provide_api_client
def list_table_lineages_cli(api_client, table_name, level):
    """
    List table lineage by table name.
    :table_name str: name of the table with 3L format. E.g catalog.schema.table

    example response:

    digraph "lineage graph of main.lineage.user_account" {
        "main.lineage.user_account" -> "main.lineage.user_transaction";
        "main.lineage.dinner_price" -> "main.lineage.price_entry","main.lineage.user_account";
    }

    Returns the specified levels of downstream/upstream

    """
    node_to_downstream = list_table_lineages_recursive_cli(api_client, table_name, level)
    click.echo(to_graph(node_to_downstream, "lineage graph of {}".format(table_name)))


@click.command(context_settings=CONTEXT_SETTINGS,
               short_help='List column lineage.')
@click.option('--table-name', required=True,
              help='Name of the table with 3L namespace')
@click.option('--column-name', required=True,
              help='Name of the column for lineage analysis')
@debug_option
@profile_option
@eat_exceptions
@provide_api_client
def list_column_lineages_cli(api_client, table_name, column_name):
    """
    List column lineage by table name and column name.
    :table_name str: name of the table with 3L format. E.g catalog.schema.table
    :column_name str: name of the column

    example response:
    {
      "downstream_cols": [
        {
          "workspace_id": 6051921418418893,
          "table_type": "TABLE",
          "catalog_name": "main",
          "table_name": "dinner_price",
          "schema_name": "lineage",
          "name": "full_menu"
        }
      ],
      "upstream_cols": [
        {
          "workspace_id": 6051921418418893,
          "table_type": "TABLE",
          "catalog_name": "main",
          "table_name": "menu",
          "schema_name": "lineage",
          "name": "app"
        },
        {
          "workspace_id": 6051921418418893,
          "table_type": "TABLE",
          "catalog_name": "main",
          "table_name": "menu",
          "schema_name": "lineage",
          "name": "desert"
        },
        {
          "workspace_id": 6051921418418893,
          "table_type": "TABLE",
          "catalog_name": "main",
          "table_name": "menu",
          "schema_name": "lineage",
          "name": "main"
        }
      ]
    }

    Returns the downstream/upstream of a given column
    """

    schemas_json = UnityCatalogApi(api_client).list_lineages_by_column(table_name, column_name)
    click.echo(mc_pretty_format(schemas_json))


@click.group()
def lineage_group():  # pragma: no cover
    pass


def register_lineage_commands(cmd_group):
    # Register deprecated "verb-noun" commands for backward compatibility.
    cmd_group.add_command(hide(list_table_lineages_cli), name='list-table-lineages')
    cmd_group.add_command(hide(list_column_lineages_cli), name='list-column-lineages')

    # Register command group.
    # Note: we deviate from the "noun-verb" pattern here because it would be awkward to have to
    # spell out "list" or "list-for". Table and column lineage is read only by definition.
    lineage_group.add_command(list_table_lineages_cli, name='table')
    lineage_group.add_command(list_column_lineages_cli, name='column')
    cmd_group.add_command(lineage_group, name='lineage')


def get_table_name(table_node):
    return "{}.{}.{}".format(
        table_node['catalog_name'],
        table_node['schema_name'],
        table_node['name']
    )


def list_table_lineages_recursive_cli(api_client, table_name, level):
    node_to_downstream = {}
    level_count = 0
    current_level = [table_name]
    next_level = []
    initial_upstream = []
    # go downstream
    while level_count < level:
        for current_table in current_level:
            if current_table in node_to_downstream:
                # skip if the table is visited before
                continue
            lineage_json = UnityCatalogApi(api_client).list_lineages_by_table(current_table)
            if level == 0:
                initial_upstream = [
                    get_table_name(
                        table_node
                    ) for table_node in lineage_json['upstream_tables']
                ] if 'upstream_tables' in lineage_json else []
            cur_downstream = [
                get_table_name(
                    table_node
                ) for table_node in lineage_json['downstream_tables']
            ] if 'downstream_tables' in lineage_json else []
            next_level.extend(cur_downstream)
            if len(cur_downstream) > 0:
                node_to_downstream[current_table] = cur_downstream
        level_count = level_count + 1
        current_level = next_level
        next_level = []
    # go upstream
    level_count = 1
    current_level = initial_upstream
    connect_upstream_tables(initial_upstream, table_name, node_to_downstream)
    next_level = []
    while level_count <= level:
        for current_table in current_level:
            lineage_json = UnityCatalogApi(api_client).list_lineages_by_table(current_table)
            upstream_of_current = [
                get_table_name(
                    table_node
                ) for table_node in lineage_json['upstream_tables']
            ] if 'upstream_tables' in lineage_json else []
            next_level.extend(upstream_of_current)
            cur_downstream = [
                get_table_name(
                    table_node
                ) for table_node in lineage_json['downstream_tables']
            ] if 'downstream_tables' in lineage_json else []
            if len(cur_downstream) > 0:
                if current_table in node_to_downstream:
                    node_to_downstream[current_table] = cur_downstream
        level_count = level_count + 1
        current_level = next_level
        next_level = []
    return node_to_downstream


def connect_upstream_tables(upstream_tables, current_table, node_to_downstream):
    """
    fill node_to_downstream dict based with give upstreams and current table
    """
    if len(upstream_tables) > 0:
        if upstream_tables not in node_to_downstream:
            node_to_downstream[upstream_tables] = [current_table]



# Databricks CLI
# Copyright 2022 Databricks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"), except
# that the use of services to which certain application programming
# interfaces (each, an "API") connect requires that the user first obtain
# a license for the use of the APIs from Databricks, Inc. ("Databricks"),
# by creating an account at www.databricks.com and agreeing to either (a)
# the Community Edition Terms of Service, (b) the Databricks Terms of
# Service, or (c) another written agreement between Licensee and Databricks
# for the use of the APIs.
#
# You may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import functools
from json import loads as json_loads

import click

from databricks_cli.click_types import JsonClickType
from databricks_cli.configure.config import provide_api_client, profile_option, debug_option
from databricks_cli.unity_catalog.api import UnityCatalogApi
from databricks_cli.unity_catalog.utils import hide, json_file_help, json_string_help, \
    mc_pretty_format
from databricks_cli.utils import eat_exceptions, CONTEXT_SETTINGS, json_cli_base, \
    merge_dicts_shallow


##############  Share Commands  ##############


@click.command(context_settings=CONTEXT_SETTINGS,
               short_help='Create a new share.')
@click.option('--name', required=True, help='Name of new share.')
@debug_option
@profile_option
@eat_exceptions
@provide_api_client
def create_share_cli(api_client, name):
    """
    Create a new share.
    """
    share_json = UnityCatalogApi(api_client).create_share(name)
    click.echo(mc_pretty_format(share_json))


@click.command(context_settings=CONTEXT_SETTINGS,
               short_help='List shares.')
@debug_option
@profile_option
@eat_exceptions
@provide_api_client
def list_shares_cli(api_client):
    """
    List shares.
    """
    shares_json = UnityCatalogApi(api_client).list_shares()
    click.echo(mc_pretty_format(shares_json))


@click.command(context_settings=CONTEXT_SETTINGS,
               short_help='Get a share.')
@click.option('--name', required=True,
              help='Name of the share to get.')
@click.option('--include-shared-data', default=True,
              help='Whether to include shared data in the response.')
@debug_option
@profile_option
@eat_exceptions
@provide_api_client
def get_share_cli(api_client, name, include_shared_data):
    """
    Get a share.
    """
    share_json = UnityCatalogApi(api_client).get_share(name, include_shared_data)
    click.echo(mc_pretty_format(share_json))


@click.command(context_settings=CONTEXT_SETTINGS,
               short_help='List permissions on a share.')
@click.option('--name', required=True,
              help='Name of the share to list permissions on.')
@debug_option
@profile_option
@eat_exceptions
@provide_api_client
def list_share_permissions_cli(api_client, name):
    """
    List permissions on a share.
    """
    perms_json = UnityCatalogApi(api_client).list_share_permissions(name)
    click.echo(mc_pretty_format(perms_json))


@click.command(context_settings=CONTEXT_SETTINGS,
               short_help='Update permissions on a share.')
@click.option('--name', required=True,
              help='Name of the share whose permissions are updated.')
@click.option('--json-file', default=None, type=click.Path(),
              help=json_file_help(method='POST', path='/shares/{name}/permissions'))
@click.option('--json', default=None, type=JsonClickType(),
              help=json_string_help(method='POST', path='/shares/{name}/permissions'))
@debug_option
@profile_option
@eat_exceptions
@provide_api_client
def update_share_permissions_cli(api_client, name, json_file, json):
    """
    Update permissions on a share.

    The public specification for the JSON request is in development.
    """
    json_cli_base(json_file, json,
                  lambda json: UnityCatalogApi(api_client).update_share_permissions(name, json))


def shared_data_object(name=None, comment=None, shared_as=None, 
                       cdf_enabled=None, partitions=None, start_version=None):
    val = {
        'data_object_type': 'TABLE'
    }
    if name is not None:
        val['name'] = name
    if comment is not None:
        val['comment'] = comment
    if shared_as is not None:
        val['shared_as'] = shared_as
    if cdf_enabled is not None:
        val['cdf_enabled'] = cdf_enabled
    if partitions is not None:
        val['partitions'] = partitions
    if start_version is not None:
        val['start_version'] = start_version
    return val


@click.command(context_settings=CONTEXT_SETTINGS,
               short_help='Update a share.')
@click.option('--name', required=True,
              help='Name of the share to update.')
@click.option('--new-name', default=None, help='New name of the share.')
@click.option('--comment', default=None, required=False,
              help='Free-form text description.')
@click.option('--owner', default=None, required=False,
              help='Owner of the share.')
# These options are hidden to encourage the usage of the new commands: add-table/update-table
@click.option('--add-table', default=None, multiple=True,
              metavar='NAME', hidden=True,
              help='Full name of table to add to share (can be specified multiple times).')
@click.option('--remove-table', default=None, multiple=True,
              metavar='NAME', hidden=True,
              help='Full name of table to remove from share (can be specified multiple times).')
@click.option('--json-file', default=None, type=click.Path(),
              help=json_file_help(method='PATCH', path='/shares/{name}'))
@click.option('--json', default=None, type=JsonClickType(),
              help=json_string_help(method='PATCH', path='/shares/{name}'))
@debug_option
@profile_option
@eat_exceptions
@provide_api_client
def update_share_cli(api_client, name, new_name, comment, owner,
                     add_table, remove_table, json_file, json):
    """
    Update a share.

    The public specification for the JSON request is in development.
    """
    if ((new_name is not None) or (comment is not None) or (owner is not None) or
        len(add_table) > 0 or len(remove_table) > 0):
        if (json_file is not None) or (json is not None):
            raise ValueError('Cannot specify JSON if any other update flags are specified')
        updates = []
        for a in add_table:
            updates.append({'action': 'ADD', 'data_object': shared_data_object(a)})
        for r in remove_table:
            updates.append({'action': 'REMOVE', 'data_object': shared_data_object(r)})
        data = {'name': new_name, 'comment': comment, 'owner': owner, 'updates': updates}
        share_json = UnityCatalogApi(api_client).update_share(name, data)
        click.echo(mc_pretty_format(share_json))
    else:
        json_cli_base(json_file, json,
                      lambda json: UnityCatalogApi(api_client).update_share(name, json))


def create_common_shared_data_object_options(f):
    @click.option('--table', required=True,
                  help='Full name of the shared table.')
    @click.option('--shared-as', default=None,
                  help='New name of the table to be shared as.')
    @click.option('--comment', default=None,
                  help='New comment of the shared table.')
    @click.option('--partitions', default=None, type=JsonClickType(),
                  help='New partition specification of the shared table represented in JSON.')
    @click.option('--cdf/--no-cdf', is_flag=True, default=None,
                  help='Toggles the change data feed for the shared table.')
    @click.option('--start-version', default=None, type=int,
                  help='Specifies the current version of the shared table.')
    @functools.wraps(f)
    def wrapper(*args, **kwargs):
        f(*args, **kwargs)
    return wrapper


@click.command(context_settings=CONTEXT_SETTINGS,
               short_help='Add a shared table.')
@click.option('--share', required=True,
              help='Name of the share to update.')
@create_common_shared_data_object_options
@click.option('--json-file', default=None, type=click.Path(),
              help="Adds a shared table based on shared data object represented in JSON file.")
@click.option('--json', default=None, type=JsonClickType(),
              help="Adds a shared table based on shared data object represented in JSON.")
@debug_option
@profile_option
@eat_exceptions
@provide_api_client
def add_share_table_cli(api_client, share, table, shared_as, comment,
                        partitions, cdf, start_version, json_file, json):
    """
    Adds a shared table.

    The public specification for the JSON request is in development.
    """
    if ((shared_as is not None) or (comment is not None) or (partitions is not None) or 
        (cdf is not None) or (start_version is not None)):
        if (json_file is not None) or (json is not None):
            raise ValueError('Cannot specify JSON if any other flags are specified')
        data = { 
            'updates': [
                {
                    'action': 'ADD',
                    'data_object': shared_data_object(
                        name=table,
                        shared_as=shared_as,
                        comment=comment,
                        cdf_enabled=cdf,
                        partitions=json_loads(partitions) if partitions is not None else None,
                        start_version=start_version,
                    )
                }
            ]
        }
        share_json = UnityCatalogApi(api_client).update_share(share, data)
        click.echo(mc_pretty_format(share_json))
    else:
        json_cli_base(json_file, json, lambda d: UnityCatalogApi(api_client).update_share(share, { 
            'updates': [
                {
                    'action': 'ADD',
                    'data_object': d,
                }
            ]
        }))


@click.command(context_settings=CONTEXT_SETTINGS,
               short_help='Update a shared table.')
@click.option('--share', required=True,
              help='Name of the share to update.')
@create_common_shared_data_object_options
@click.option('--json-file', default=None, type=click.Path(),
              help="Updates the shared table to shared data object represented in JSON file.")
@click.option('--json', default=None, type=JsonClickType(),
              help="Updates the shared table to shared data object represented in JSON.")
@debug_option
@profile_option
@eat_exceptions
@provide_api_client
def update_share_table_cli(api_client, share, table, shared_as, comment,
                           partitions, cdf, start_version, json_file, json):
    """
    Updates a shared table.

    The public specification for the JSON request is in development.
    """
    if ((shared_as is not None) or (comment is not None) or (partitions is not None) or 
        (cdf is not None) or (start_version is not None)):
        if (json_file is not None) or (json is not None):
            raise ValueError('Cannot specify JSON if any other flags are specified')
        data = { 
            'updates': [
                {
                    'action': 'UPDATE',
                    'data_object': shared_data_object(
                        name=table,
                        shared_as=shared_as,
                        comment=comment,
                        cdf_enabled=cdf,
                        partitions=json_loads(partitions) if partitions is not None else None,
                        start_version=start_version,
                    )
                }
            ]
        }
        share_json = UnityCatalogApi(api_client).update_share(share, data)
        click.echo(mc_pretty_format(share_json))
    else:
        json_cli_base(json_file, json, lambda d: UnityCatalogApi(api_client).update_share(share, { 
            'updates': [
                {
                    'action': 'UPDATE',
                    'data_object': d,
                }
            ]
        }))


@click.command(context_settings=CONTEXT_SETTINGS,
               short_help='Update a shared table.')
@click.option('--share', required=True,
              help='Name of the share to update.')
@click.option('--table', default=None,
              help='Full name of the table to update from share.')
@click.option('--shared-as', default=None,
              help='New name of the table inside the share.')
@click.option('--json-file', default=None, type=click.Path(),
              help="Removes the shared table based on shared data object represented in JSON file.")
@click.option('--json', default=None, type=JsonClickType(),
              help="Removes the shared table based on shared data object represented in JSON.")
@debug_option
@profile_option
@eat_exceptions
@provide_api_client
def remove_share_table_cli(api_client, share, table, shared_as, json_file, json):
    """
    Removes a shared table either by table name or the shared-as table name.

    The public specification for the JSON request is in development.
    """
    if table is not None and shared_as is not None:
        raise ValueError("You can only pass in either --table or --shared_as and not both.")

    if table is not None or shared_as is not None:
        if (json_file is not None) or (json is not None):
            raise ValueError('Cannot specify JSON if any other flags are specified')
        data = { 
            'updates': [
                {
                    'action': 'REMOVE',
                    'data_object': shared_data_object(
                        name=table,
                        shared_as=shared_as,
                    )
                }
            ]
        }
        share_json = UnityCatalogApi(api_client).update_share(share, data)
        click.echo(mc_pretty_format(share_json))
    else:
        json_cli_base(json_file, json, lambda d: UnityCatalogApi(api_client).update_share(share, { 
            'updates': [
                {
                    'action': 'REMOVE',
                    'data_object': d,
                }
            ]
        }))


@click.command(context_settings=CONTEXT_SETTINGS,
               short_help='Delete a share.')
@click.option('--name', required=True,
              help='Name of the share to delete.')
@debug_option
@profile_option
@eat_exceptions
@provide_api_client
def delete_share_cli(api_client, name):
    """
    Delete a share.
    """
    UnityCatalogApi(api_client).delete_share(name)


##############  Recipient Commands  ##############

def parse_recipient_custom_properties(custom_property_list):
    custom_properties = []
    for property_str in custom_property_list:
        tokens = property_str.split('=', 2)
        if len(tokens) != 2:
            raise ValueError('Invalid format for property. '
                             + 'The format should be <key>=<value>.')
        custom_properties.append({"key": tokens[0], "value": tokens[1]})
    return custom_properties    


@click.command(context_settings=CONTEXT_SETTINGS,
               short_help='Create a new recipient.')
@click.option('--name', required=True, help='Name of new recipient.')
@click.option('--comment', default=None, required=False,
              help='Free-form text description.')
@click.option('--sharing-id', default=None, required=False,
              help='The sharing identifier provided by the data recipient offline.')
@click.option('--allowed-ip-address', default=None, required=False, multiple=True,
              help=(
                  'IP address in CIDR notation that is allowed to use delta sharing. '
                  '(can be specified multiple times).'))
@click.option('--property', 'custom_property', default=None, required=False, multiple=True,
              help=(
                  'Properties of the recipient. Key and value should be provided '
                  'at the same time separated by an equal sign. '
                  'Example: --property country=US.'))
@debug_option
@profile_option
@eat_exceptions
@provide_api_client
def create_recipient_cli(api_client, name, comment, sharing_id,
                         allowed_ip_address, custom_property):
    """
    Create a new recipient.
    """
    recipient_json = UnityCatalogApi(api_client).create_recipient(
        name, comment, sharing_id,
        allowed_ip_address, parse_recipient_custom_properties(custom_property))
    click.echo(mc_pretty_format(recipient_json))


@click.command(context_settings=CONTEXT_SETTINGS,
               short_help='List recipients.')
@debug_option
@profile_option
@eat_exceptions
@provide_api_client
def list_recipients_cli(api_client):
    """
    List recipients.
    """
    recipients_json = UnityCatalogApi(api_client).list_recipients()
    click.echo(mc_pretty_format(recipients_json))


@click.command(context_settings=CONTEXT_SETTINGS,
               short_help='Get a recipient.')
@click.option('--name', required=True,
              help='Name of the recipient to get.')
@debug_option
@profile_option
@eat_exceptions
@provide_api_client
def get_recipient_cli(api_client, name):
    """
    Get a recipient.
    """
    recipient_json = UnityCatalogApi(api_client).get_recipient(name)
    click.echo(mc_pretty_format(recipient_json))


@click.command(context_settings=CONTEXT_SETTINGS,
               short_help='Update a recipient.')
@click.option('--name', required=True,
              help='Name of the recipient who needs to be updated.')
@click.option('--new-name', default=None, help='New name of the recipient.')
@click.option('--comment', default=None, required=False,
              help='Free-form text description.')
@click.option('--owner', default=None, required=False,
              help='Owner of the recipient.')
@click.option('--allowed-ip-address', default=None, required=False, multiple=True,
              help=(
                  'IP address in CIDR notation that is allowed to use delta sharing '
                  '(can be specified multiple times). Specify a single empty string to disable '
                  'IP allowlist.'))
@click.option('--property', 'custom_property', default=None, required=False, multiple=True,
              help=(
                  'Properties of the recipient. Key and value should be provided '
                  'at the same time separated by an equal sign. '
                  'Example: --property country=US. '
                  'Specify a single empty string to remove all properties.'))
@click.option('--json-file', default=None, type=click.Path(),
              help=json_file_help(method='PATCH', path='/recipients/{name}'))
@click.option('--json', default=None, type=JsonClickType(),
              help=json_string_help(method='PATCH', path='/recipients/{name}'))
@debug_option
@profile_option
@eat_exceptions
@provide_api_client
def update_recipient_cli(api_client, name, new_name, comment, owner,
                         allowed_ip_address, custom_property, json_file, json):
    """
    Update a recipient.

    The public specification for the JSON request is in development.
    """
    if ((new_name is not None) or (comment is not None) or (owner is not None) or
        len(allowed_ip_address) or len(custom_property) > 0):
        if (json_file is not None) or (json is not None):
            raise ValueError('Cannot specify JSON if any other update flags are specified')
        data = {'name': new_name, 'comment': comment, 'owner': owner}
        if len(allowed_ip_address) > 0:
            data['ip_access_list'] = {}
            if len(allowed_ip_address) != 1 or allowed_ip_address[0] != '':
                data['ip_access_list']['allowed_ip_addresses'] = allowed_ip_address
        if len(custom_property) > 0:
            data['properties_kvpairs'] = {}
            if len(custom_property) != 1 or custom_property[0] != '':
                data['properties_kvpairs']['properties'] = parse_recipient_custom_properties(
                    custom_property)
        recipient_json = UnityCatalogApi(api_client).update_recipient(name, data)
        click.echo(mc_pretty_format(recipient_json))
    else:
        json_cli_base(json_file, json,
                      lambda json: UnityCatalogApi(api_client).update_recipient(name, json))


@click.command(context_settings=CONTEXT_SETTINGS,
               short_help='Rotate token for the recipient.')
@click.option('--name', required=True, help='Name of new recipient.')
@click.option('--existing-token-expire-in-seconds', default=None, required=False,
              type=int,
              help='Expire the existing token in number of seconds from now,' +
                   ' 0 to expire it immediately.')
@debug_option
@profile_option
@eat_exceptions
@provide_api_client
def rotate_recipient_token_cli(api_client, name, existing_token_expire_in_seconds):
    """
    Rotate recipient token.
    """
    recipient_json = \
        UnityCatalogApi(api_client).rotate_recipient_token(name, existing_token_expire_in_seconds)
    click.echo(mc_pretty_format(recipient_json))


@click.command(context_settings=CONTEXT_SETTINGS,
               short_help='List share permissions of a recipient.')
@click.option('--name', required=True,
              help='Name of the recipient.')
@debug_option
@profile_option
@eat_exceptions
@provide_api_client
def list_recipient_permissions_cli(api_client, name):
    """
    List a recipient's share permissions.
    """
    recipient_json = UnityCatalogApi(api_client).get_recipient_share_permissions(name)
    click.echo(mc_pretty_format(recipient_json))


@click.command(context_settings=CONTEXT_SETTINGS,
               short_help='Delete a recipient.')
@click.option('--name', required=True,
              help='Name of the recipient to delete.')
@debug_option
@profile_option
@eat_exceptions
@provide_api_client
def delete_recipient_cli(api_client, name):
    """
    Delete a recipient.
    """
    UnityCatalogApi(api_client).delete_recipient(name)


##############  Provider Commands  ##############

@click.command(context_settings=CONTEXT_SETTINGS,
               short_help='Create a provider.')
@click.option('--name', required=True, help='Name of the new provider.')
@click.option('--comment', default=None, required=False,
              help='Free-form text description.')
@click.option('--recipient-profile-json-file', default=None, required=False, type=click.Path(),
              help='File containing recipient profile in JSON format.')
@click.option('--recipient-profile-json', default=None, required=False, type=JsonClickType(),
              help='JSON string containing recipient profile.')
@debug_option
@profile_option
@eat_exceptions
@provide_api_client
def create_provider_cli(api_client, name, comment, recipient_profile_json_file,
                        recipient_profile_json):
    """
    Create a provider.

    The public specification for the JSON request is in development.
    """
    json_cli_base(recipient_profile_json_file, recipient_profile_json,
                  lambda json: UnityCatalogApi(api_client).create_provider(name, comment, json),
                  error_msg='Either --recipient-profile-json-file or ' +
                  '--recipient-profile-json should be provided')


@click.command(context_settings=CONTEXT_SETTINGS,
               short_help='List providers.')
@debug_option
@profile_option
@eat_exceptions
@provide_api_client
def list_providers_cli(api_client):
    """
    List providers.
    """
    proviers_json = UnityCatalogApi(api_client).list_providers()
    click.echo(mc_pretty_format(proviers_json))


@click.command(context_settings=CONTEXT_SETTINGS,
               short_help='Get a provider.')
@click.option('--name', required=True,
              help='Name of the provider to get.')
@debug_option
@profile_option
@eat_exceptions
@provide_api_client
def get_provider_cli(api_client, name):
    """
    Get a provider.
    """
    provier_json = UnityCatalogApi(api_client).get_provider(name)
    click.echo(mc_pretty_format(provier_json))


@click.command(context_settings=CONTEXT_SETTINGS,
               short_help='Update a provider.')
@click.option('--name', required=True, help='Name of the provider to update.')
@click.option('--new-name', default=None, help='New name of the provider.')
@click.option('--comment', default=None, required=False,
              help='Free-form text description.')
@click.option('--owner', default=None, required=False,
              help='Owner of the provider.')
@click.option('--recipient-profile-json-file', default=None, type=click.Path(),
              help='File containing recipient profile in JSON format.')
@click.option('--recipient-profile-json', default=None, type=JsonClickType(),
              help='JSON string containing recipient profile.')
@click.option('--json-file', default=None, type=click.Path(),
              help=json_file_help(method='PATCH', path='/providers/{name}'))
@click.option('--json', default=None, type=JsonClickType(),
              help=json_string_help(method='PATCH', path='/providers/{name}'))
@debug_option
@profile_option
@eat_exceptions
@provide_api_client
def update_provider_cli(api_client, name, new_name, comment, owner, recipient_profile_json_file,
                        recipient_profile_json, json_file, json):
    """
    Update a provider.

    The public specification for the JSON request is in development.
    """
    if ((new_name is not None) or (comment is not None) or (owner is not None) or
        (recipient_profile_json_file is not None) or (recipient_profile_json) is not None):
        if (json_file is not None) or (json is not None):
            raise ValueError('Cannot specify JSON if any other update flags are specified')
        data = {'name': new_name, 'comment': comment, 'owner': owner}

        if (recipient_profile_json_file is None) and (recipient_profile_json is None):
            provider_json = UnityCatalogApi(api_client).update_provider(name, provider_spec=data)
            click.echo(mc_pretty_format(provider_json))
        else:
            json_cli_base(recipient_profile_json_file, recipient_profile_json,
                          lambda profile_json: UnityCatalogApi(api_client).update_provider(
                              name,
                              provider_spec=merge_dicts_shallow(
                                  data,
                                  {'recipient_profile_str': mc_pretty_format(profile_json)})),
                          error_msg='Either --recipient-profile-json-file or ' +
                          '--recipient-profile-json should be provided')
    else:
        json_cli_base(json_file, json,
                      lambda json: UnityCatalogApi(api_client).update_provider(
                          name, provider_spec=json))
            

@click.command(context_settings=CONTEXT_SETTINGS,
               short_help='List shares of a provider.')
@click.option('--name', required=True,
              help='Name of the provider.')
@debug_option
@profile_option
@eat_exceptions
@provide_api_client
def list_provider_shares_cli(api_client, name):
    """
    List a provider's shares.
    """
    shares_json = UnityCatalogApi(api_client).list_provider_shares(name)
    click.echo(mc_pretty_format(shares_json))


@click.command(context_settings=CONTEXT_SETTINGS,
               short_help='Delete a provider.')
@click.option('--name', required=True,
              help='Name of the provider to delete.')
@debug_option
@profile_option
@eat_exceptions
@provide_api_client
def delete_provider_cli(api_client, name):
    """
    Delete a provider.
    """
    UnityCatalogApi(api_client).delete_provider(name)


@click.group()
def shares_group():  # pragma: no cover
    pass


def register_shares_commands(cmd_group):
    # Register deprecated "verb-noun" commands for backward compatibility.
    cmd_group.add_command(hide(create_share_cli), name='create-share')
    cmd_group.add_command(hide(list_shares_cli), name='list-shares')
    cmd_group.add_command(hide(get_share_cli), name='get-share')
    cmd_group.add_command(hide(update_share_cli), name='update-share')
    cmd_group.add_command(hide(delete_share_cli), name='delete-share')
    cmd_group.add_command(hide(list_share_permissions_cli), name='list-share-permissions')
    cmd_group.add_command(hide(update_share_permissions_cli), name='update-share-permissions')

    # Register command group.
    shares_group.add_command(create_share_cli, name='create')
    shares_group.add_command(list_shares_cli, name='list')
    shares_group.add_command(get_share_cli, name='get')
    shares_group.add_command(update_share_cli, name='update')
    shares_group.add_command(add_share_table_cli, name='add-table')
    shares_group.add_command(update_share_table_cli, name='update-table')
    shares_group.add_command(remove_share_table_cli, name='remove-table')
    shares_group.add_command(delete_share_cli, name='delete')
    shares_group.add_command(list_share_permissions_cli, name='list-permissions')
    shares_group.add_command(update_share_permissions_cli, name='update-permissions')
    cmd_group.add_command(shares_group, name='shares')


@click.group()
def recipients_group():  # pragma: no cover
    pass


def register_recipients_commands(cmd_group):
    # Register deprecated "verb-noun" commands for backward compatibility.
    cmd_group.add_command(hide(create_recipient_cli), name='create-recipient')
    cmd_group.add_command(hide(list_recipients_cli), name='list-recipients')
    cmd_group.add_command(hide(get_recipient_cli), name='get-recipient')
    cmd_group.add_command(hide(update_recipient_cli), name='update-recipient')
    cmd_group.add_command(hide(rotate_recipient_token_cli), name='rotate-recipient-token')
    cmd_group.add_command(hide(list_recipient_permissions_cli), name='list-recipient-permissions')
    cmd_group.add_command(hide(delete_recipient_cli), name='delete-recipient')

    # Register command group.
    recipients_group.add_command(create_recipient_cli, name='create')
    recipients_group.add_command(list_recipients_cli, name='list')
    recipients_group.add_command(get_recipient_cli, name='get')
    recipients_group.add_command(update_recipient_cli, name='update')
    recipients_group.add_command(rotate_recipient_token_cli, name='rotate-token')
    recipients_group.add_command(list_recipient_permissions_cli, name='list-permissions')
    recipients_group.add_command(delete_recipient_cli, name='delete')
    cmd_group.add_command(recipients_group, name='recipients')


@click.group()
def providers_group():  # pragma: no cover
    pass


def register_providers_commands(cmd_group):
    # Register deprecated "verb-noun" commands for backward compatibility.
    cmd_group.add_command(hide(create_provider_cli), name='create-provider')
    cmd_group.add_command(hide(list_providers_cli), name='list-providers')
    cmd_group.add_command(hide(get_provider_cli), name='get-provider')
    cmd_group.add_command(hide(update_provider_cli), name='update-provider')
    cmd_group.add_command(hide(delete_provider_cli), name='delete-provider')
    cmd_group.add_command(hide(list_provider_shares_cli), name='list-provider-shares')

    # Register command group.
    providers_group.add_command(create_provider_cli, name='create')
    providers_group.add_command(list_providers_cli, name='list')
    providers_group.add_command(get_provider_cli, name='get')
    providers_group.add_command(update_provider_cli, name='update')
    providers_group.add_command(delete_provider_cli, name='delete')
    providers_group.add_command(list_provider_shares_cli, name='list-shares')
    cmd_group.add_command(providers_group, name='providers')


def register_delta_sharing_commands(cmd_group):
    register_shares_commands(cmd_group)
    register_recipients_commands(cmd_group)
    register_providers_commands(cmd_group)



# Databricks CLI
# Copyright 2018 Databricks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"), except
# that the use of services to which certain application programming
# interfaces (each, an "API") connect requires that the user first obtain
# a license for the use of the APIs from Databricks, Inc. ("Databricks"),
# by creating an account at www.databricks.com and agreeing to either (a)
# the Community Edition Terms of Service, (b) the Databricks Terms of
# Service, or (c) another written agreement between Licensee and Databricks
# for the use of the APIs.
#
# You may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os
import json

import click

from databricks_cli.utils import eat_exceptions, CONTEXT_SETTINGS
from databricks_cli.version import print_version_callback, version
from databricks_cli.configure.config import provide_api_client, profile_option, debug_option
from databricks_cli.stack.api import StackApi

DEBUG_MODE = True


def _generate_stack_status_path(stack_path):
    """
    Given a path to the stack configuration template JSON file, generates a path to where the
    deployment status JSON will be stored after successful deployment of the stack.

    :param stack_path: Path to the stack config template JSON file
    :return: The path to the stack status file.

    >>> self._generate_stack_status_path('./stack.json')
    './stack.deployed.json'
    """
    stack_status_insert = 'deployed'
    stack_path_split = stack_path.split('.')
    stack_path_split.insert(-1, stack_status_insert)
    return '.'.join(stack_path_split)


def _load_json(path):
    """
    Parse a json file to a readable dict format.
    Returns an empty dictionary if the path doesn't exist.

    :param path: File path of the JSON stack configuration template.
    :return: dict of parsed JSON stack config template.
    """
    stack_conf = {}
    if os.path.exists(path):
        with open(path, 'r') as f:
            stack_conf = json.load(f)
    return stack_conf


def _save_json(path, data):
    """
    Writes data to a JSON file.

    :param path: Path of JSON file.
    :param data: dict- data that wants to by written to JSON file
    :return: None
    """
    with open(path, 'w') as f:
        json.dump(data, f, indent=2, sort_keys=True)


@click.command(context_settings=CONTEXT_SETTINGS,
               short_help='Deploy a stack given a JSON configuration of the stack')
@click.argument('config_path', type=click.Path(exists=True), required=True)
@click.option('--overwrite', '-o', is_flag=True, default=False, show_default=True,
              help='Include to overwrite existing workspace notebooks and dbfs files')
@debug_option
@profile_option
@eat_exceptions
@provide_api_client
def deploy(api_client, config_path, **kwargs):
    """
    Deploy a stack to the databricks workspace given a JSON stack configuration template.

    After deployment, a stack status will be saves at <config_file_name>.deployed.json. Please
    do not edit or move the file as it is generated through the CLI and is used for future
    deployments of the stack. If you run into errors with the stack status at deployment,
    please delete the stack status file and try the deployment again. If the problem persists,
    please raise a Github issue on the Databricks CLI repository at
    https://www.github.com/databricks/databricks-cli/issues
    """
    click.echo('#' * 80)
    click.echo('Deploying stack at: {} with options: {}'.format(config_path, kwargs))
    stack_config = _load_json(config_path)
    status_path = _generate_stack_status_path(config_path)
    stack_status = _load_json(status_path)
    config_dir = os.path.dirname(os.path.abspath(config_path))
    cli_dir = os.getcwd()
    os.chdir(config_dir)  # Switch current working directory to where json config is stored
    new_stack_status = StackApi(api_client).deploy(stack_config, stack_status, **kwargs)
    click.echo('#' * 80)
    os.chdir(cli_dir)
    click.echo("Saving stack status to {}".format(status_path))
    _save_json(status_path, new_stack_status)
    click.echo('Note: The stack status file is an automatically generated file that is'
               ' depended on for the databricks stack CLI to function correctly.'
               ' Please do not edit the file.')
    click.echo('#' * 80)


@click.command(context_settings=CONTEXT_SETTINGS,
               short_help='Download stack notebooks given a JSON configuration of the stack')
@click.argument('config_path', type=click.Path(exists=True), required=True)
@click.option('--overwrite', '-o', is_flag=True, help='Include to overwrite existing'
                                                      ' notebooks in the local filesystem.')
@debug_option
@profile_option
@eat_exceptions
@provide_api_client
def download(api_client, config_path, **kwargs):
    """
    Download workspace notebooks of a stack to the local filesystem given a JSON stack
    configuration template.
    """
    click.echo('#' * 80)
    click.echo('Downloading stack at: {} with options: {}'.format(config_path, kwargs))
    stack_config = _load_json(config_path)
    config_dir = os.path.dirname(os.path.abspath(config_path))
    cli_dir = os.getcwd()
    os.chdir(config_dir)  # Switch current working directory to where json config is stored
    StackApi(api_client).download(stack_config, **kwargs)
    os.chdir(cli_dir)
    click.echo('#' * 80)


@click.group(context_settings=CONTEXT_SETTINGS,
             short_help='[Beta] Utility to deploy and download Databricks resource stacks.')
@click.option('--version', '-v', is_flag=True, callback=print_version_callback,
              expose_value=False, is_eager=True, help=version)
@debug_option
@profile_option
def stack_group():  # pragma: no cover
    """
    [Beta] Utility to deploy and download Databricks resource stacks.
    """
    pass


stack_group.add_command(deploy, name='deploy')
stack_group.add_command(download, name='download')



# Databricks CLI
# Copyright 2017 Databricks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"), except
# that the use of services to which certain application programming
# interfaces (each, an "API") connect requires that the user first obtain
# a license for the use of the APIs from Databricks, Inc. ("Databricks"),
# by creating an account at www.databricks.com and agreeing to either (a)
# the Community Edition Terms of Service, (b) the Databricks Terms of
# Service, or (c) another written agreement between Licensee and Databricks
# for the use of the APIs.
#
# You may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import time
from datetime import datetime
from json import loads as json_loads

import click
from tabulate import tabulate

from databricks_cli.click_types import OutputClickType, JsonClickType, ClusterIdClickType, \
    OneOfOption
from databricks_cli.clusters.api import ClusterApi
from databricks_cli.configure.config import provide_api_client, profile_option, debug_option
from databricks_cli.utils import eat_exceptions, CONTEXT_SETTINGS, pretty_format, json_cli_base, \
    truncate_string, CLUSTER_OPTIONS
from databricks_cli.version import print_version_callback, version


@click.command(context_settings=CONTEXT_SETTINGS)
@click.option('--json-file', default=None, type=click.Path(),
              help='File containing JSON request to POST to /api/2.0/clusters/create.')
@click.option('--json', default=None, type=JsonClickType(),
              help=JsonClickType.help('/api/2.0/clusters/create'))
@debug_option
@profile_option
@eat_exceptions
@provide_api_client
def create_cli(api_client, json_file, json):
    """
    Creates a Databricks cluster.

    The specification for the request json can be found at
    https://docs.databricks.com/api/latest/clusters.html#create
    """
    json_cli_base(json_file, json, lambda json: ClusterApi(api_client).create_cluster(json))


@click.command(context_settings=CONTEXT_SETTINGS)
@click.option('--json-file', default=None, type=click.Path(),
              help='File containing JSON request to POST to /api/2.0/clusters/edit.')
@click.option('--json', default=None, type=JsonClickType(),
              help=JsonClickType.help('/api/2.0/clusters/edit'))
@debug_option
@profile_option
@eat_exceptions
@provide_api_client
def edit_cli(api_client, json_file, json):
    """
    Edits a Databricks cluster.

    The specification for the request json can be found at
    https://docs.databricks.com/api/latest/clusters.html#edit
    """
    if not bool(json_file) ^ bool(json):
        raise RuntimeError('Either --json-file or --json should be provided')
    if json_file:
        with open(json_file, 'r') as f:
            json = f.read()
    deser_json = json_loads(json)
    ClusterApi(api_client).edit_cluster(deser_json)


@click.command(context_settings=CONTEXT_SETTINGS,
               short_help='Starts a terminated Databricks cluster given its ID.')
@click.option('--cluster-id', required=True, type=ClusterIdClickType(),
              help=ClusterIdClickType.help)
@debug_option
@profile_option
@eat_exceptions
@provide_api_client
def start_cli(api_client, cluster_id):
    """
    Starts a terminated Databricks cluster given its ID.

    If the cluster is not currently in a TERMINATED state, nothing will happen.

    """
    ClusterApi(api_client).start_cluster(cluster_id)


@click.command(context_settings=CONTEXT_SETTINGS)
@click.option('--cluster-id', required=True, type=ClusterIdClickType(),
              help=ClusterIdClickType.help)
@debug_option
@profile_option
@provide_api_client
@eat_exceptions
def restart_cli(api_client, cluster_id):
    """
    Restarts a Databricks cluster given its ID.

    If the cluster is not currently in a RUNNING state, nothing will happen
    """
    ClusterApi(api_client).restart_cluster(cluster_id)


@click.command(context_settings=CONTEXT_SETTINGS)
@click.option('--cluster-id', required=True, type=ClusterIdClickType(),
              help=ClusterIdClickType.help)
@click.option('--num-workers', required=True, type=click.INT,
              help='Number of workers')
@debug_option
@profile_option
@provide_api_client
@eat_exceptions
def resize_cli(api_client, cluster_id, num_workers):
    """Resizes a Databricks cluster given its ID.

    Provide a `--num-workers` parameter to indicate the new cluster size.

    If the cluster is not currently in a RUNNING state, this will cause an
    error to occur.
    """
    ClusterApi(api_client).resize_cluster(cluster_id, num_workers)


@click.command(context_settings=CONTEXT_SETTINGS)
@click.option('--cluster-id', required=True, type=ClusterIdClickType(),
              help=ClusterIdClickType.help)
@debug_option
@profile_option
@eat_exceptions
@provide_api_client
def delete_cli(api_client, cluster_id):
    """
    Removes a Databricks cluster given its ID.

    The cluster is removed asynchronously. Once the deletion has completed,
    the cluster will be in a TERMINATED state. If the cluster is already in
    a TERMINATING or TERMINATED state, nothing will happen.

    Use ``databricks clusters get --cluster-id CLUSTER_ID`` to check termination states.
    """
    ClusterApi(api_client).delete_cluster(cluster_id)


@click.command(context_settings=CONTEXT_SETTINGS)
@click.option('--cluster-id', cls=OneOfOption, one_of=CLUSTER_OPTIONS,
              type=ClusterIdClickType(), default=None, help=ClusterIdClickType.help)
@click.option('--cluster-name', cls=OneOfOption, one_of=CLUSTER_OPTIONS,
              type=ClusterIdClickType(), default=None, help=ClusterIdClickType.help)
@debug_option
@profile_option
@eat_exceptions
@provide_api_client
def get_cli(api_client, cluster_id, cluster_name):
    """
    Retrieves metadata about a cluster.
    """
    if cluster_id:
        cluster = ClusterApi(api_client).get_cluster(cluster_id)
    elif cluster_name:
        cluster = ClusterApi(api_client).get_cluster_by_name(cluster_name)
    else:
        raise RuntimeError('cluster_name and cluster_id were empty?')

    click.echo(pretty_format(cluster))


def _clusters_to_table(clusters_json):
    ret = []
    for c in clusters_json.get('clusters', []):
        ret.append((c['cluster_id'], truncate_string(c['cluster_name']), c['state']))
    return ret


def _cluster_events_to_table(events_json):
    ret = []
    for event in events_json.get('events', []):
        timestamp = event['timestamp'] / 1000
        timezone = time.tzname[time.localtime(timestamp).tm_isdst]
        formatted_time = "%s %s" % (
            datetime.fromtimestamp(event['timestamp'] / 1000.).strftime('%Y-%m-%d %H:%M:%S'),
            timezone)
        ret.append((formatted_time, event['type'], event['details']))
    return ret


@click.command(context_settings=CONTEXT_SETTINGS,
               short_help='Lists active and recently terminated clusters.')
@click.option('--output', default=None, help=OutputClickType.help, type=OutputClickType())
@debug_option
@profile_option
@eat_exceptions
@provide_api_client
def list_cli(api_client, output):
    """
    Lists active and recently terminated clusters.

    Returns information about all currently active clusters, and up
    to 100 most recently terminated clusters in the past 7 days.

    By default the output format will be a human readable table with the following fields

      - Cluster ID

      - Cluster name

      - Cluster state
    """
    clusters_json = ClusterApi(api_client).list_clusters()
    if OutputClickType.is_json(output):
        click.echo(pretty_format(clusters_json))
    else:
        click.echo(tabulate(_clusters_to_table(clusters_json), tablefmt='plain'))


@click.command(context_settings=CONTEXT_SETTINGS)
@debug_option
@profile_option
@eat_exceptions
@provide_api_client
def list_zones_cli(api_client):
    """
    Lists zones where clusters can be created.

    The output format is specified in
    https://docs.databricks.com/api/latest/clusters.html#list-zones
    """
    click.echo(pretty_format(ClusterApi(api_client).list_zones()))


@click.command(context_settings=CONTEXT_SETTINGS,
               short_help='Lists possible node types for a cluster.')
@debug_option
@profile_option
@eat_exceptions
@provide_api_client
def list_node_types_cli(api_client):
    """
    Lists possible node types for a cluster.

    The output format is specified in
    https://docs.databricks.com/api/latest/clusters.html#list-node-types
    """
    click.echo(pretty_format(ClusterApi(api_client).list_node_types()))


@click.command(context_settings=CONTEXT_SETTINGS)
@debug_option
@profile_option
@eat_exceptions
@provide_api_client
def spark_versions_cli(api_client):
    """
    Lists possible Databricks Runtime versions for a cluster.

    The output format is specified in
    https://docs.databricks.com/api/latest/clusters.html#spark-versions
    """
    click.echo(pretty_format(ClusterApi(api_client).spark_versions()))


@click.command(context_settings=CONTEXT_SETTINGS)
@click.option('--cluster-id', required=True, type=ClusterIdClickType(),
              help=ClusterIdClickType.help)
@debug_option
@profile_option
@eat_exceptions
@provide_api_client
def permanent_delete_cli(api_client, cluster_id):
    """
    Permanently deletes a Spark cluster.

    If the cluster is running, it is terminated and its resources are asynchronously removed.
    If the cluster is terminated, then it is immediately removed.
    """
    ClusterApi(api_client).permanent_delete(cluster_id)


@click.command(context_settings=CONTEXT_SETTINGS)
@click.option('--cluster-id', required=True, type=ClusterIdClickType(),
              help=ClusterIdClickType.help)
@click.option('--start-time', required=False, default=None,
              help="The start time in epoch milliseconds. If unprovided, returns events starting "
                   "from the beginning of time.")
@click.option('--end-time', required=False, default=None,
              help="The end time in epoch milliseconds. If unprovided, returns events up to the "
                   "current time")
@click.option('--order', required=False, default=None,
              help="The order to list events in; either ASC or DESC. Defaults to DESC "
                   "(most recent first).")
@click.option('--event-type', required=False, default=None,
              help="An event types to filter on (specify multiple event types by passing "
                   "the --event-type option multiple times). If empty, all event types "
                   "are returned.", multiple=True)
@click.option('--offset', required=False, default=None,
              help="The offset in the result set. Defaults to 0 (no offset). When an offset is "
                   "specified and the results are requested in descending order, the end_time "
                   "field is required.")
@click.option('--limit', required=False, default=None,
              help="The maximum number of events to include in a page of events. Defaults to 50, "
                   "and maximum allowed value is 500.")
@click.option('--output', default=None, help=OutputClickType.help, type=OutputClickType())
@debug_option
@profile_option
@eat_exceptions
@provide_api_client
def cluster_events_cli(api_client, cluster_id, start_time, end_time, order, event_type, offset,
                       limit, output):
    """
    Gets events for a Spark cluster.
    """
    events_json = ClusterApi(api_client).get_events(
        cluster_id=cluster_id, start_time=start_time, end_time=end_time, order=order,
        event_types=event_type, offset=offset, limit=limit)
    if OutputClickType.is_json(output):
        click.echo(pretty_format(events_json))
    else:
        click.echo(tabulate(_cluster_events_to_table(events_json), tablefmt='plain'))


@click.group(context_settings=CONTEXT_SETTINGS,
             short_help='Utility to interact with Databricks clusters.')
@click.option('--version', '-v', is_flag=True, callback=print_version_callback,
              expose_value=False, is_eager=True, help=version)
@debug_option
@profile_option
@eat_exceptions
def clusters_group():  # pragma: no cover
    """
    Utility to interact with Databricks clusters.
    """
    pass


clusters_group.add_command(create_cli, name='create')
clusters_group.add_command(edit_cli, name='edit')
clusters_group.add_command(start_cli, name='start')
clusters_group.add_command(restart_cli, name='restart')
clusters_group.add_command(resize_cli, name='resize')
clusters_group.add_command(delete_cli, name='delete')
clusters_group.add_command(get_cli, name='get')
clusters_group.add_command(list_cli, name='list')
clusters_group.add_command(list_zones_cli, name='list-zones')
clusters_group.add_command(list_node_types_cli, name='list-node-types')
clusters_group.add_command(spark_versions_cli, name='spark-versions')
clusters_group.add_command(permanent_delete_cli, name='permanent-delete')
clusters_group.add_command(cluster_events_cli, name='events')
